import re
import unicodedata
from collections import OrderedDict, defaultdict
from datetime import date, datetime, timedelta
from sys import platform
from time import ctime, sleep, time
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from urllib.parse import urljoin, urlparse

import dateutil.parser
import urwid
from bs4 import BeautifulSoup
from bs4.element import NavigableString
from tzlocal import get_localzone
from urwid_readline import ReadlineEdit

from zulipterminal.config.keys import (
    is_command_key,
    keys_for_command,
    primary_key_for_command,
)
from zulipterminal.config.symbols import (
    MESSAGE_CONTENT_MARKER,
    MESSAGE_HEADER_DIVIDER,
    QUOTED_TEXT_MARKER,
    STREAM_MARKER_INVALID,
    STREAM_MARKER_PRIVATE,
    STREAM_MARKER_PUBLIC,
    STREAM_TOPIC_SEPARATOR,
    TIME_MENTION_MARKER,
)
from zulipterminal.helper import (
    Message,
    asynch,
    format_string,
    get_unused_fence,
    match_emoji,
    match_group,
    match_stream,
    match_topics,
    match_user,
)
from zulipterminal.server_url import near_message_url
from zulipterminal.ui_tools.buttons import EditModeButton
from zulipterminal.ui_tools.tables import render_table
from zulipterminal.urwid_types import urwid_Size


class WriteBox(urwid.Pile):
    def __init__(self, view: Any) -> None:
        super().__init__(self.main_view(True))
        self.model = view.model
        self.view = view

        # If editing a message, its id - otherwise None
        self.msg_edit_id = None  # type: Optional[int]
        # Determines if the message body (content) can be edited
        self.msg_body_edit_enabled = True

        self.is_in_typeahead_mode = False

        # Set to int for stream box only
        self.stream_id = None  # type: Optional[int]

        # Used in PM and stream boxes
        # (empty list implies PM box empty, or not initialized)
        # * prioritizes autocomplete in message body
        # * updates server on PM typing events
        self.recipient_user_ids = []  # type: List[int]

        # Private message recipient text entry
        self.to_write_box = None  # None if stream-box or not initialized

        # For tracking sending typing status updates
        self.send_next_typing_update = datetime.now()
        self.last_key_update = datetime.now()
        self.idle_status_tracking = False

        # Constant indices into self.contents
        # (CONTAINER=vertical, HEADER/MESSAGE=horizontal)
        self.FOCUS_CONTAINER_HEADER = 0
        self.FOCUS_HEADER_BOX_RECIPIENT = 0
        self.FOCUS_HEADER_BOX_STREAM = 1
        self.FOCUS_HEADER_BOX_TOPIC = 3
        self.FOCUS_HEADER_BOX_EDIT = 4
        self.FOCUS_CONTAINER_MESSAGE = 1
        self.FOCUS_MESSAGE_BOX_BODY = 0
        # These are included to allow improved clarity
        # FIXME: These elements don't acquire focus; replace prefix & in above?
        self.FOCUS_HEADER_PREFIX_STREAM = 0
        self.FOCUS_HEADER_PREFIX_TOPIC = 2

    def main_view(self, new: bool) -> Any:
        if new:
            return []
        else:
            self.contents.clear()

    def set_editor_mode(self) -> None:
        self.view.controller.enter_editor_mode_with(self)

    def send_stop_typing_status(self) -> None:
        # Send 'stop' updates only for PM narrows.
        if self.to_write_box and self.recipient_user_ids:
            self.model.send_typing_status_by_user_ids(
                self.recipient_user_ids,
                status='stop'
            )
            self.send_next_typing_update = datetime.now()
            self.idle_status_tracking = False

    def private_box_view(self, *, email: str='',
                         recipient_user_ids: Optional[List[int]]=None) -> None:
        # Neither or both arguments should be set
        assert ((email != '' and recipient_user_ids is not None)
                or (email == '' and recipient_user_ids is None))

        self.set_editor_mode()
        if recipient_user_ids:
            self.recipient_user_ids = recipient_user_ids
        else:
            self.recipient_user_ids = []

        self.send_next_typing_update = datetime.now()
        self.to_write_box = ReadlineEdit("To: ", edit_text=email)
        self.msg_write_box = ReadlineEdit(multiline=True)
        self.msg_write_box.enable_autocomplete(
            func=self.generic_autocomplete,
            key=primary_key_for_command('AUTOCOMPLETE'),
            key_reverse=primary_key_for_command('AUTOCOMPLETE_REVERSE')
        )
        self.header_write_box = urwid.Columns([self.to_write_box])
        header_line_box = urwid.LineBox(
            self.header_write_box,
            tlcorner='━', tline='━', trcorner='━', lline='',
            blcorner='─', bline='─', brcorner='─', rline=''
        )
        self.contents = [
            (header_line_box, self.options()),
            (self.msg_write_box, self.options()),
        ]
        self.focus_position = self.FOCUS_CONTAINER_MESSAGE

        # Typing status is sent in regular intervals to limit the number of
        # notifications sent. Idleness should also prompt a notification.
        # Refer to https://zulip.com/api/set-typing-status for the protocol
        # on typing notifications sent by clients.
        TYPING_STARTED_WAIT_PERIOD = 10
        TYPING_STOPPED_WAIT_PERIOD = 5

        start_period_delta = timedelta(seconds=TYPING_STARTED_WAIT_PERIOD)
        stop_period_delta = timedelta(seconds=TYPING_STOPPED_WAIT_PERIOD)

        def on_type_send_status(edit: object, new_edit_text: str) -> None:
            if new_edit_text:
                self.last_key_update = datetime.now()
                if self.last_key_update > self.send_next_typing_update:
                    self.model.send_typing_status_by_user_ids(
                                self.recipient_user_ids, status='start')
                    self.send_next_typing_update += start_period_delta
                    # Initiate tracker function only if it isn't already
                    # initiated.
                    if not self.idle_status_tracking:
                        self.idle_status_tracking = True
                        track_idleness_and_update_status()

        @asynch
        def track_idleness_and_update_status() -> None:
            while datetime.now() < self.last_key_update + stop_period_delta:
                idle_check_time = (self.last_key_update
                                   + stop_period_delta
                                   - datetime.now())
                sleep(idle_check_time.total_seconds())
            self.send_stop_typing_status()

        urwid.connect_signal(self.msg_write_box, 'change', on_type_send_status)

    def stream_box_view(self, stream_id: int, caption: str='', title: str='',
                        ) -> None:
        self.set_editor_mode()
        self.stream_id = stream_id
        self.recipient_user_ids = self.model.get_other_subscribers_in_stream(
                                            stream_id=stream_id)
        self.to_write_box = None
        self.msg_write_box = ReadlineEdit(multiline=True)
        self.msg_write_box.enable_autocomplete(
            func=self.generic_autocomplete,
            key=primary_key_for_command('AUTOCOMPLETE'),
            key_reverse=primary_key_for_command('AUTOCOMPLETE_REVERSE')
        )
        self.stream_write_box = ReadlineEdit(edit_text=caption)
        self.stream_write_box.enable_autocomplete(
            func=self._stream_box_autocomplete,
            key=primary_key_for_command('AUTOCOMPLETE'),
            key_reverse=primary_key_for_command('AUTOCOMPLETE_REVERSE')
        )
        self.stream_write_box.set_completer_delims("")

        self.title_write_box = ReadlineEdit(edit_text=title)
        self.title_write_box.enable_autocomplete(
            func=self._topic_box_autocomplete,
            key=primary_key_for_command('AUTOCOMPLETE'),
            key_reverse=primary_key_for_command('AUTOCOMPLETE_REVERSE')
        )
        self.title_write_box.set_completer_delims("")

        stream_marker = STREAM_MARKER_PUBLIC
        color = None
        if caption:
            color = self.model.stream_dict[self.stream_id]['color']
            if self.model.stream_dict[self.stream_id]['invite_only']:
                stream_marker = STREAM_MARKER_PRIVATE
        self.header_write_box = urwid.Columns([
            ('pack', urwid.Text((color, STREAM_MARKER_PUBLIC))),
            self.stream_write_box,
            ('pack', urwid.Text(STREAM_TOPIC_SEPARATOR)),
            self.title_write_box], dividechars=1)
        header_line_box = urwid.LineBox(
            self.header_write_box,
            tlcorner='━', tline='━', trcorner='━', lline='',
            blcorner='─', bline='─', brcorner='─', rline=''
        )
        write_box = [
            (header_line_box, self.options()),
            (self.msg_write_box, self.options()),
        ]
        self.contents = write_box

        # Use and set a callback to set the stream marker
        self._set_stream_write_box_style(None, caption)
        urwid.connect_signal(self.stream_write_box, 'change',
                             self._set_stream_write_box_style)

    def stream_box_edit_view(self, stream_id: int, caption: str='',
                             title: str='') -> None:
        self.stream_box_view(stream_id, caption, title)
        self.edit_mode_button = EditModeButton(self.model.controller, 20)

        self.header_write_box.widget_list.append(self.edit_mode_button)

    def _set_stream_write_box_style(self, widget: ReadlineEdit,
                                    new_text: str) -> None:
        # FIXME: Refactor when we have ~ Model.is_private_stream
        stream_marker = STREAM_MARKER_INVALID
        color = 'general_bar'
        if self.model.is_valid_stream(new_text):
            stream = self.model.stream_dict[
                     self.model.stream_id_from_name(new_text)]
            if stream['invite_only']:
                stream_marker = STREAM_MARKER_PRIVATE
            else:
                stream_marker = STREAM_MARKER_PUBLIC
            color = stream['color']
        (self.header_write_box[self.FOCUS_HEADER_PREFIX_STREAM]
         .set_text((color, stream_marker)))

    def _topic_box_autocomplete(self, text: str, state: Optional[int]
                                ) -> Optional[str]:
        topic_names = self.model.topics_in_stream(self.stream_id)

        topic_typeaheads = match_topics(topic_names, text)

        # Typeaheads and suggestions are the same.
        return self._process_typeaheads(topic_typeaheads, state,
                                        topic_typeaheads)

    def _stream_box_autocomplete(self, text: str, state: Optional[int]
                                 ) -> Optional[str]:
        streams_list = self.view.pinned_streams + self.view.unpinned_streams
        streams = [stream['name']
                   for stream in streams_list]

        # match_streams takes stream names and typeaheads,
        # but we don't have typeaheads here.
        # FIXME: Refactor match_stream
        stream_data = list(zip(streams, streams))
        matched_streams = match_stream(stream_data, text,
                                       self.view.pinned_streams)

        # matched_streams[0] and matched_streams[1] contains the same data.
        return self._process_typeaheads(matched_streams[0], state,
                                        matched_streams[1])

    def generic_autocomplete(self, text: str, state: Optional[int]
                             ) -> Optional[str]:
        autocomplete_map = OrderedDict([
                ('@_', self.autocomplete_users),
                ('@_**', self.autocomplete_users),
                ('@', self.autocomplete_mentions),
                ('@*', self.autocomplete_groups),
                ('@**', self.autocomplete_users),
                ('#', self.autocomplete_streams),
                ('#**', self.autocomplete_streams),
                (':', self.autocomplete_emojis),
            ])

        # Look in a reverse order to find the last autocomplete prefix used in
        # the text. For instance, if text='@#example', use '#' as the prefix.
        # FIXME: Mentions can actually start with '#', and streams with
        #        anything; this implementation simply chooses the right-most
        #        match of the longest length
        prefix_indices = {
            prefix: text.rfind(prefix)
            for prefix in autocomplete_map
        }
        found_prefix_indices = {
            prefix: index
            for prefix, index in prefix_indices.items()
            if index > -1
        }
        # Return text if it doesn't have any of the autocomplete prefixes.
        if not found_prefix_indices:
            return text

        # Use latest longest matching prefix (so @_ wins vs @)
        prefix_index = max(found_prefix_indices.values())
        prefix = max(
            (len(prefix), prefix)
            for prefix, index in found_prefix_indices.items()
            if index == prefix_index
        )[1]
        autocomplete_func = autocomplete_map[prefix]

        # NOTE: The following block only executes if any of the autocomplete
        # prefixes exist.
        typeaheads, suggestions = (
            autocomplete_func(text[prefix_index:], prefix)
        )

        typeahead = self._process_typeaheads(typeaheads, state,
                                             suggestions)
        if typeahead:
            typeahead = text[:prefix_index] + typeahead
        return typeahead

    def _process_typeaheads(self, typeaheads: List[str], state: Optional[int],
                            suggestions: List[str]) -> Optional[str]:
        num_suggestions = 10
        fewer_typeaheads = typeaheads[:num_suggestions]
        reduced_suggestions = suggestions[:num_suggestions]
        is_truncated = len(fewer_typeaheads) != len(typeaheads)

        if (state is not None and state < len(fewer_typeaheads)
                and state >= -len(fewer_typeaheads)):
            typeahead = fewer_typeaheads[state]  # type: Optional[str]
        else:
            typeahead = None
            state = None
        self.is_in_typeahead_mode = True
        self.view.set_typeahead_footer(reduced_suggestions,
                                       state, is_truncated)
        return typeahead

    def autocomplete_mentions(self, text: str, prefix_string: str
                              ) -> Tuple[List[str], List[str]]:
        # Handles user mentions (@ mentions and silent mentions)
        # and group mentions.

        user_typeahead, user_names = self.autocomplete_users(
            text, prefix_string
        )
        group_typeahead, groups = self.autocomplete_groups(
            text, prefix_string
        )

        combined_typeahead = user_typeahead + group_typeahead
        combined_names = user_names + groups

        return combined_typeahead, combined_names

    def autocomplete_users(self, text: str, prefix_string: str
                           ) -> Tuple[List[str], List[str]]:
        users_list = self.view.users
        matching_users = [user
                          for user in users_list
                          if match_user(user, text[len(prefix_string):])]
        matching_ids = set([user['user_id'] for user in matching_users])
        matching_recipient_ids = (set(self.recipient_user_ids)
                                  & set(matching_ids))
        # Display subscribed users/recipients first.
        sorted_matching_users = sorted(matching_users,
                                       key=lambda user: user['user_id']
                                       in matching_recipient_ids,
                                       reverse=True)

        user_names = [user['full_name'] for user in sorted_matching_users]
        extra_prefix = "{}{}".format(
            '*' if prefix_string[-1] != '*' else '',
            '*' if prefix_string[-2:] != '**' else '',
        )
        user_typeahead = format_string(user_names,
                                       prefix_string + extra_prefix + '{}**')

        return user_typeahead, user_names

    def autocomplete_groups(self, text: str, prefix_string: str
                            ) -> Tuple[List[str], List[str]]:
        prefix_length = len(prefix_string)
        groups = [group_name
                  for group_name in self.model.user_group_names
                  if match_group(group_name, text[prefix_length:])]

        extra_prefix = '*' if prefix_string[-1] != '*' else ''
        group_typeahead = format_string(groups,
                                        prefix_string + extra_prefix + '{}*')
        return group_typeahead, groups

    def autocomplete_streams(self, text: str, prefix_string: str
                             ) -> Tuple[List[str], List[str]]:
        streams_list = self.view.pinned_streams + self.view.unpinned_streams
        streams = [stream['name']
                   for stream in streams_list]
        stream_typeahead = format_string(streams, '#**{}**')
        stream_data = list(zip(stream_typeahead, streams))

        prefix_length = len(prefix_string)

        matched_data = match_stream(stream_data, text[prefix_length:],
                                    self.view.pinned_streams)
        return matched_data

    def autocomplete_emojis(self, text: str, prefix_string: str
                            ) -> Tuple[List[str], List[str]]:
        emoji_list = list(self.model.active_emoji_data.keys())
        emojis = [emoji
                  for emoji in emoji_list
                  if match_emoji(emoji, text[1:])]
        emoji_typeahead = format_string(emojis, ':{}:')

        return emoji_typeahead, emojis

    def keypress(self, size: urwid_Size, key: str) -> Optional[str]:
        if self.is_in_typeahead_mode:
            if not (is_command_key('AUTOCOMPLETE', key)
                    or is_command_key('AUTOCOMPLETE_REVERSE', key)):
                # set default footer when done with autocomplete
                self.is_in_typeahead_mode = False
                self.view.set_footer_text()

        if is_command_key('SEND_MESSAGE', key):
            self.send_stop_typing_status()
            if not self.to_write_box:
                if re.fullmatch(r'\s*', self.title_write_box.edit_text):
                    topic = '(no topic)'
                else:
                    topic = self.title_write_box.edit_text

                if self.msg_edit_id:
                    args = dict(message_id=self.msg_edit_id,
                                topic=topic,
                                propagate_mode=self.edit_mode_button.mode)
                    if self.msg_body_edit_enabled:
                        args['content'] = self.msg_write_box.edit_text

                    success = self.model.update_stream_message(**args)
                else:
                    success = self.model.send_stream_message(
                        stream=self.stream_write_box.edit_text,
                        topic=topic,
                        content=self.msg_write_box.edit_text
                    )
            else:
                if self.msg_edit_id:
                    success = self.model.update_private_message(
                        content=self.msg_write_box.edit_text,
                        msg_id=self.msg_edit_id,
                    )
                else:
                    recipient_emails = [email.strip() for email in
                                        self.to_write_box.edit_text.split(',')]
                    success = self.model.send_private_message(
                        recipients=recipient_emails,
                        content=self.msg_write_box.edit_text
                    )
            if success:
                self.msg_write_box.edit_text = ''
                if self.msg_edit_id:
                    self.msg_edit_id = None
                    self.keypress(size, 'esc')
        elif is_command_key('GO_BACK', key):
            self.msg_edit_id = None
            self.msg_body_edit_enabled = True
            self.send_stop_typing_status()
            self.view.controller.exit_editor_mode()
            self.main_view(False)
            self.view.middle_column.set_focus('body')
        elif is_command_key('SAVE_AS_DRAFT', key):
            if not self.msg_edit_id:
                if self.to_write_box:
                    message = Message(
                            display_recipient=self.to_write_box.edit_text,
                            content=self.msg_write_box.edit_text,
                            type='private',
                    )
                elif self.stream_id:
                    message = Message(
                        display_recipient=self.stream_write_box.edit_text,
                        content=self.msg_write_box.edit_text,
                        subject=self.title_write_box.edit_text,
                        stream_id=self.stream_id,
                        type='stream',
                    )
                saved_draft = self.model.session_draft_message()
                if not saved_draft:
                    self.model.save_draft(message)
                elif message != saved_draft:
                    self.view.controller.save_draft_confirmation_popup(message)
        elif is_command_key('CYCLE_COMPOSE_FOCUS', key):
            if len(self.contents) == 0:
                return key
            header = self.header_write_box
            # toggle focus position
            if self.focus_position == self.FOCUS_CONTAINER_HEADER:
                if self.to_write_box is None:
                    if header.focus_col == self.FOCUS_HEADER_BOX_STREAM:
                        stream_name = (header[self.FOCUS_HEADER_BOX_STREAM]
                                       .edit_text)
                        if not self.model.is_valid_stream(stream_name):
                            invalid_stream_error = (
                                'Invalid stream name.'
                                ' Use {} or {} to autocomplete.'
                                .format(primary_key_for_command(
                                            'AUTOCOMPLETE'
                                        ),
                                        primary_key_for_command(
                                            'AUTOCOMPLETE'
                                            '_REVERSE'
                                        ))
                            )
                            self.view.set_footer_text(invalid_stream_error, 3)
                            return key
                        user_ids = self.model.get_other_subscribers_in_stream(
                                                    stream_name=stream_name)
                        self.recipient_user_ids = user_ids
                        self.stream_id = self.model.stream_id_from_name(
                                                                stream_name)

                        header.focus_col = self.FOCUS_HEADER_BOX_TOPIC
                        return key
                    elif (header.focus_col == self.FOCUS_HEADER_BOX_TOPIC
                          and self.msg_edit_id):
                        header.focus_col = self.FOCUS_HEADER_BOX_EDIT
                        return key
                    elif header.focus_col == self.FOCUS_HEADER_BOX_EDIT:
                        if self.msg_body_edit_enabled:
                            header.focus_col = self.FOCUS_HEADER_BOX_STREAM
                            self.focus_position = self.FOCUS_CONTAINER_MESSAGE
                        else:
                            header.focus_col = self.FOCUS_HEADER_BOX_TOPIC
                        return key
                    else:
                        header.focus_col = self.FOCUS_HEADER_BOX_STREAM
                else:
                    recipient_box = header[self.FOCUS_HEADER_BOX_RECIPIENT]
                    recipient_emails = [email.strip() for email in
                                        recipient_box.edit_text.split(',')]
                    invalid_emails = self.model.get_invalid_recipient_emails(
                                                              recipient_emails)
                    if invalid_emails:
                        invalid_emails_error = ('Invalid recipient(s) - '
                                                + ', '.join(invalid_emails))
                        self.view.set_footer_text(invalid_emails_error, 3)
                        return key
                    users = self.model.user_dict
                    self.recipient_user_ids = [users[email]['user_id']
                                               for email in recipient_emails]

            if not self.msg_body_edit_enabled:
                return key
            if self.focus_position == self.FOCUS_CONTAINER_HEADER:
                self.focus_position = self.FOCUS_CONTAINER_MESSAGE
            else:
                self.focus_position = self.FOCUS_CONTAINER_HEADER
            if self.to_write_box is None:
                header.focus_col = self.FOCUS_HEADER_BOX_STREAM
            else:
                header.focus_col = self.FOCUS_HEADER_BOX_RECIPIENT

        key = super().keypress(size, key)
        return key


class MessageBox(urwid.Pile):
    # type of last_message is Optional[Message], but needs refactoring
    def __init__(self, message: Message, model: Any,
                 last_message: Any) -> None:
        self.model = model
        self.message = message
        self.stream_name = ''
        self.stream_id = None  # type: Union[int, None]
        self.topic_name = ''
        self.email = ''
        self.user_id = None  # type: Union[int, None]
        self.message_links = (
            OrderedDict()
        )   # type: OrderedDict[str, Tuple[str, int, bool]]
        self.time_mentions = list()  # type: List[Tuple[str, str]]
        self.last_message = last_message
        # if this is the first message
        if self.last_message is None:
            self.last_message = defaultdict(dict)

        if self.message['type'] == 'stream':
            self.stream_name = self.message['display_recipient']
            self.stream_id = self.message['stream_id']
            self.topic_name = self.message['subject']
        elif self.message['type'] == 'private':
            self.email = self.message['sender_email']
            self.user_id = self.message['sender_id']
        else:
            raise RuntimeError("Invalid message type")

        if self.message['type'] == 'private':
            if self._is_private_message_to_self():
                recipient = self.message['display_recipient'][0]
                self.recipients_names = recipient['full_name']
                self.recipients_emails = self.model.user_email
                self.recipient_ids = self.model.user_id
            else:
                self.recipients_names = ', '.join(list(
                            recipient['full_name']
                            for recipient in self.message['display_recipient']
                            if recipient['email'] != self.model.user_email
                        ))
                self.recipients_emails = ', '.join(list(
                            recipient['email']
                            for recipient in self.message['display_recipient']
                            if recipient['email'] != self.model.user_email
                        ))
                self.recipient_ids = [recipient['id'] for recipient
                                      in self.message['display_recipient']
                                      if recipient['id'] != self.model.user_id]

        # mouse_event helper variable
        self.displaying_selection_hint = False

        super().__init__(self.main_view())

    def _time_for_message(self, message: Message) -> str:
        return ctime(message['timestamp'])[:-8]

    def need_recipient_header(self) -> bool:
        # Prevent redundant information in recipient bar
        if (len(self.model.narrow) == 1
                and self.model.narrow[0][0] == 'pm_with'):
            return False
        if (len(self.model.narrow) == 2
                and self.model.narrow[1][0] == 'topic'):
            return False

        last_msg = self.last_message
        if self.message['type'] == 'stream':
            return not (
                last_msg['type'] == 'stream'
                and self.topic_name == last_msg['subject']
                and self.stream_name == last_msg['display_recipient']
            )
        elif self.message['type'] == 'private':
            recipient_ids = [{recipient['id']
                              for recipient in message['display_recipient']
                              if 'id' in recipient}
                             for message in (self.message, last_msg)
                             if 'display_recipient' in message]
            return not (
                len(recipient_ids) == 2
                and recipient_ids[0] == recipient_ids[1]
                and last_msg['type'] == 'private'
            )
        else:
            raise RuntimeError("Invalid message type")

    def _is_private_message_to_self(self) -> bool:
        recipient_list = self.message['display_recipient']
        return (len(recipient_list) == 1
                and recipient_list[0]['email'] == self.model.user_email)

    def stream_header(self) -> Any:
        color = self.model.stream_dict[self.stream_id]['color']
        bar_color = 's' + color
        stream_title_markup = ('bar', [
            (bar_color, '{} {} '.format(self.stream_name,
                                        STREAM_TOPIC_SEPARATOR)),
            ('title', ' {}'.format(self.topic_name))
        ])
        stream_title = urwid.Text(stream_title_markup)
        header = urwid.Columns([
            ('pack', stream_title),
            (1, urwid.Text((color, ' '))),
            urwid.AttrWrap(urwid.Divider(MESSAGE_HEADER_DIVIDER),
                           color),
        ])
        header.markup = stream_title_markup
        return header

    def private_header(self) -> Any:
        title_markup = ('header', [
            ('general_narrow', 'You and '),
            ('general_narrow', self.recipients_names)
        ])
        title = urwid.Text(title_markup)
        header = urwid.Columns([
            ('pack', title),
            (1, urwid.Text(('general_bar', ' '))),
            urwid.AttrWrap(urwid.Divider(MESSAGE_HEADER_DIVIDER),
                           'general_bar'),
        ])
        header.markup = title_markup
        return header

    def top_header_bar(self, message_view: Any) -> Any:
        if self.message['type'] == 'stream':
            return message_view.stream_header()
        else:
            return message_view.private_header()

    def top_search_bar(self) -> Any:
        curr_narrow = self.model.narrow
        is_search_narrow = self.model.is_search_narrow()
        if is_search_narrow:
            curr_narrow = [sub_narrow for sub_narrow in curr_narrow
                           if sub_narrow[0] != 'search']
        else:
            self.model.controller.view.search_box.text_box.set_edit_text("")
        if curr_narrow == []:
            text_to_fill = 'All messages'
        elif len(curr_narrow) == 1 and curr_narrow[0][1] == 'private':
            text_to_fill = 'All private messages'
        elif len(curr_narrow) == 1 and curr_narrow[0][1] == 'starred':
            text_to_fill = 'Starred messages'
        elif len(curr_narrow) == 1 and curr_narrow[0][1] == 'mentioned':
            text_to_fill = 'Mentions'
        elif self.message['type'] == 'stream':
            bar_color = self.model.stream_dict[self.stream_id]['color']
            bar_color = 's' + bar_color
            if len(curr_narrow) == 2 and curr_narrow[1][0] == 'topic':
                text_to_fill = ('bar', [  # type: ignore
                    (bar_color, '{}'.format(self.stream_name)),
                    (bar_color, ': topic narrow')
                ])
            else:
                text_to_fill = ('bar', [  # type: ignore
                    (bar_color, '{}'.format(self.stream_name))
                ])
        elif len(curr_narrow) == 1 and len(curr_narrow[0][1].split(",")) > 1:
            text_to_fill = 'Group private conversation'
        else:
            text_to_fill = 'Private conversation'

        if is_search_narrow:
            title_markup = ('header', [
                ('general_narrow', text_to_fill),
                (None, ' '),
                ('filter_results', 'Search Results')
            ])
        else:
            title_markup = ('header', [
                ('general_narrow', text_to_fill)
            ])
        title = urwid.Text(title_markup)
        header = urwid.AttrWrap(title, 'bar')
        header.text_to_fill = text_to_fill
        header.markup = title_markup
        return header

    def reactions_view(self, reactions: List[Dict[str, Any]]) -> Any:
        if not reactions:
            return ''
        try:
            reaction_stats = defaultdict(set)
            for reaction in reactions:
                user_id = int(reaction['user'].get('id', -1))
                if user_id == -1:
                    user_id = int(reaction['user']['user_id'])
                reaction_stats[reaction['emoji_name']].add(user_id)

            sorted_stats = sorted(
                (reaction, count)
                for reaction, count in reaction_stats.items()
            )

            my_user_id = self.model.user_id
            reaction_texts = [
                ('reaction_mine' if my_user_id in ids else 'reaction',
                 ':{}: {}'.format(reaction, len(ids)))
                for reaction, ids in sorted_stats
            ]

            spaced_reaction_texts = [
                entry
                for pair in zip(reaction_texts,
                                ' ' * len(reaction_texts))
                for entry in pair
            ]
            return urwid.Padding(
                urwid.Text(spaced_reaction_texts),
                align='left', width=('relative', 90), left=25, min_width=50)
        except Exception:
            return ''

    # Use quotes as a workaround for OrderedDict typing issue.
    # See https://github.com/python/mypy/issues/6904.
    def footlinks_view(
        self, message_links: 'OrderedDict[str, Tuple[str, int, bool]]',
    ) -> Any:
        # Return if footlinks are disabled by the user.
        if not self.model.controller.footlinks_enabled:
            return None

        footlinks = []
        for link, (text, index, show_footlink) in message_links.items():
            if not show_footlink:
                continue

            footlinks.extend([
                ('msg_link_index', '{}:'.format(index)),
                ' ',
                ('msg_link', link),
                '\n',
            ])

        if not footlinks:
            return None

        footlinks[-1] = footlinks[-1][:-1]  # Remove the last newline.
        return urwid.Padding(urwid.Text(footlinks, wrap='ellipsis'),
                             align='left', left=8, width=('relative', 100),
                             min_width=10, right=2)

    def soup2markup(self, soup: Any, **state: Any) -> List[Any]:
        # Ensure a string is provided, in case the soup finds none
        # This could occur if eg. an image is removed or not shown
        markup = ['']  # type: List[Union[str, Tuple[Optional[str], Any]]]
        if soup is None:  # This is not iterable, so return promptly
            return markup
        unrendered_tags = {  # In pairs of 'tag_name': 'text'
            # TODO: Some of these could be implemented
            'br': '',  # No indicator of absence
            'hr': 'RULER',
            'img': 'IMAGE',
        }
        unrendered_div_classes = {  # In pairs of 'div_class': 'text'
            # TODO: Support embedded content & twitter preview?
            'message_embed': 'EMBEDDED CONTENT',
            'inline-preview-twitter': 'TWITTER PREVIEW',
            'message_inline_ref': '',  # Duplicate of other content
            'message_inline_image': '',  # Duplicate of other content
        }
        unrendered_template = '[{} NOT RENDERED]'
        for element in soup:
            if isinstance(element, NavigableString):
                # NORMAL STRINGS
                if (hasattr(self, 'bq_len') and element == '\n'
                        and self.bq_len > 0):
                    self.bq_len -= 1
                    continue
                markup.append(element)
            elif (element.name == 'div' and element.attrs
                  and any(cls in element.attrs.get('class', [])
                          for cls in unrendered_div_classes)):
                # UNRENDERED DIV CLASSES
                matching_class = (set(unrendered_div_classes)
                                  & set(element.attrs.get('class')))
                text = unrendered_div_classes[matching_class.pop()]
                if text:
                    markup.append(unrendered_template.format(text))
            elif (element.name == 'img'
                  and element.attrs.get('class', []) == ['emoji']):
                # CUSTOM EMOJIS AND ZULIP_EXTRA_EMOJI
                emoji_name = element.attrs.get('title', [])
                markup.append(('msg_emoji', ":" + emoji_name + ":"))
            elif element.name in unrendered_tags:
                # UNRENDERED SIMPLE TAGS
                text = unrendered_tags[element.name]
                if text:
                    markup.append(unrendered_template.format(text))
            elif element.name in ('p', 'del'):
                # PARAGRAPH, STRIKE-THROUGH
                markup.extend(self.soup2markup(element))
            elif (element.name == 'span' and element.attrs
                  and 'emoji' in element.attrs.get('class', [])):
                # EMOJI
                markup.append(('msg_emoji', element.text))
            elif (element.name == 'span' and element.attrs
                  and ('katex-display' in element.attrs.get('class', [])
                       or 'katex' in element.attrs.get('class', []))):
                # MATH TEXT
                markup.append(element.text)
            elif (element.name == 'span' and element.attrs
                  and ('user-group-mention' in element.attrs.get('class', [])
                       or 'user-mention' in element.attrs.get('class', []))):
                # USER MENTIONS & USER-GROUP MENTIONS
                markup.append(('msg_mention', element.text))
            elif element.name == 'a':
                # LINKS
                # Use rstrip to avoid anomalies and edge cases like
                # https://google.com vs https://google.com/.
                link = element.attrs['href'].rstrip('/')
                text = element.img['src'] if element.img else element.text
                text = text.rstrip('/')

                parsed_link = urlparse(link)
                if not parsed_link.scheme:  # => relative link
                    # Prepend org url to convert it to an absolute link
                    link = urljoin(self.model.server_url, link)

                text = text if text else link

                show_footlink = True
                # Only use the last segment if the text is redundant.
                # NOTE: The 'without scheme' excerpt is to deal with the case
                # where a user puts a link without any scheme and the server
                # uses http as the default scheme but keeps the text as-is.
                # For instance, see how example.com/some/path becomes
                # <a href="http://example.com">example.com/some/path</a>.
                link_without_scheme, text_without_scheme = [
                    data.split('://')[1] if '://' in data else data
                    for data in [link, text]
                ]   # Split on '://' is for cases where text == link.
                if link_without_scheme == text_without_scheme:
                    last_segment = text.split('/')[-1]
                    if '.' in last_segment:
                        new_text = last_segment  # Filename.
                    elif text.startswith(self.model.server_url):
                        # Relative URL.
                        new_text = text.split(self.model.server_url)[-1]
                    else:
                        new_text = (
                            parsed_link.netloc if parsed_link.netloc
                            else text.split('/')[0]
                        )  # Domain name.
                    if new_text != text_without_scheme:
                        text = new_text
                    else:
                        # Do not show as a footlink as the text is sufficient
                        # to represent the link.
                        show_footlink = False

                # Detect duplicate links to save screen real estate.
                if link not in self.message_links:
                    self.message_links[link] = (
                        text, len(self.message_links) + 1, show_footlink
                    )
                else:
                    # Append the text if its link already exist with a
                    # different text.
                    saved_text, saved_link_index, saved_footlink_status = (
                        self.message_links[link]
                    )
                    if saved_text != text:
                        self.message_links[link] = (
                            '{}, {}'.format(saved_text, text),
                            saved_link_index,
                            show_footlink or saved_footlink_status,
                        )

                markup.extend([
                    ('msg_link', text),
                    ' ',
                    ('msg_link_index',
                     '[{}]'.format(self.message_links[link][1])),
                ])
            elif element.name == 'blockquote':
                # BLOCKQUOTE TEXT
                markup.append((
                    'msg_quote', self.soup2markup(element)
                ))
            elif element.name == 'code':
                # CODE (INLINE?)
                markup.append((
                    'msg_code', element.text
                ))
            elif (element.name == 'div' and element.attrs
                    and 'codehilite' in element.attrs.get('class', [])):
                # CODE (BLOCK?)
                markup.append((
                    'msg_code', element.text
                ))
            elif element.name in ('strong', 'em'):
                # BOLD & ITALIC
                markup.append(('msg_bold', element.text))
            elif element.name in ('ul', 'ol'):
                # LISTS (UL & OL)
                for part in element.contents:
                    if part == '\n':
                        part.replace_with('')

                if 'indent_level' not in state:
                    state['indent_level'] = 1
                    state['list_start'] = True
                else:
                    state['indent_level'] += 1
                    state['list_start'] = False
                if element.name == 'ol':
                    start_number = int(element.attrs.get('start', 1))
                    state['list_index'] = start_number
                    markup.extend(self.soup2markup(element, **state))
                    del state['list_index']  # reset at end of this list
                else:
                    if 'list_index' in state:
                        del state['list_index']  # this is unordered
                    markup.extend(self.soup2markup(element, **state))
                del state['indent_level']  # reset indents after any list
            elif element.name == 'li':
                # LIST ITEMS (LI)
                for part in element.contents:
                    if part == '\n':
                        part.replace_with('')
                if not state.get('list_start', False):
                    markup.append('\n')

                indent = state.get('indent_level', 1)
                if 'list_index' in state:
                    markup.append('{}{}. '.format('  ' * indent,
                                                  state['list_index']))
                    state['list_index'] += 1
                else:
                    chars = [
                            '\N{BULLET}',
                            '\N{RING OPERATOR}',    # small hollow
                            '\N{HYPHEN}',
                    ]
                    markup.append('{}{} '.format('  ' * indent,
                                                 chars[(indent - 1) % 3]))
                state['list_start'] = False
                markup.extend(self.soup2markup(element, **state))
            elif element.name == 'table':
                markup.extend(render_table(element))
            elif element.name == 'time':
                # New in feature level 16, server version 3.0.
                # Render time in current user's local time zone.
                timestamp = element.get('datetime')

                # This should not happen. Regardless, we are interested in
                # debugging and reporting it to zulip/zulip if it does.
                assert timestamp is not None, 'Could not find datetime attr'

                utc_time = dateutil.parser.parse(timestamp)
                local_time = utc_time.astimezone(get_localzone())
                # TODO: Address 12-hour format support with application-wide
                # support for different formats.
                time_string = local_time.strftime('%a, %b %-d %Y, %-H:%M (%Z)')
                markup.append((
                    'msg_time',
                    ' {} {} '.format(TIME_MENTION_MARKER, time_string)
                ))

                source_text = (
                    'Original text was {}'.format(element.text.strip())
                )
                self.time_mentions.append((time_string, source_text))
            else:
                markup.extend(self.soup2markup(element))
        return markup

    def main_view(self) -> List[Any]:

        # Recipient Header
        if self.need_recipient_header():
            if self.message['type'] == 'stream':
                recipient_header = self.stream_header()
            else:
                recipient_header = self.private_header()
        else:
            recipient_header = None

        # Content Header
        message = {
            key: {
                'is_starred': 'starred' in msg['flags'],
                'author': (msg['sender_full_name']
                           if 'sender_full_name' in msg else None),
                'time': (self._time_for_message(msg)
                         if 'timestamp' in msg else None),
                'datetime': (datetime.fromtimestamp(msg['timestamp'])
                             if 'timestamp' in msg else None),
            }
            for key, msg in dict(this=self.message,
                                 last=self.last_message).items()
        }
        different = {  # How this message differs from the previous one
            'recipients': recipient_header is not None,
            'author': message['this']['author'] != message['last']['author'],
            '24h': (message['last']['datetime'] is not None
                    and ((message['this']['datetime']
                          - message['last']['datetime'])
                         .days)),
            'timestamp': (
                message['last']['time'] is not None
                and message['this']['time'] != message['last']['time']
            ),
            'star_status': (
                message['this']['is_starred'] != message['last']['is_starred']
            ),
        }
        any_differences = any(different.values())

        if any_differences:  # Construct content_header, if needed
            TextType = Dict[str, Tuple[Optional[str], str]]
            text_keys = ('author', 'star', 'time')
            text = {key: (None, ' ') for key in text_keys}  # type: TextType

            if any(different[key] for key in ('recipients', 'author', '24h')):
                text['author'] = ('name', message['this']['author'])
            if message['this']['is_starred']:
                text['star'] = ('starred', "*")
            if any(different[key]
                   for key in ('recipients', 'author', 'timestamp')):
                this_year = date.today().year
                msg_year = message['this']['datetime'].year
                if this_year != msg_year:
                    text['time'] = (
                        'time',
                        '{} - {}'.format(msg_year, message['this']['time'])
                    )
                else:
                    text['time'] = ('time', message['this']['time'])

            content_header = urwid.Columns([
                ('weight', 10, urwid.Text(text['author'])),
                (23, urwid.Text(text['time'], align='right')),
                (1, urwid.Text(text['star'], align='right')),
                ], dividechars=1)
        else:
            content_header = None

        # If the message contains '/me' emote then replace it with
        # sender's full name and show it in bold.
        if self.message['is_me_message']:
            self.message['content'] = self.message['content'].replace(
                '/me',
                '<strong>' + self.message['sender_full_name'] + '</strong>', 1)

        # Transform raw message content into markup (As needed by urwid.Text)
        content = self.transform_content()

        if self.message['id'] in self.model.index['edited_messages']:
            edited_label_size = 7
            left_padding = 1
        else:
            edited_label_size = 0
            left_padding = 8

        content = urwid.Padding(
            urwid.Columns([
                (edited_label_size,
                 urwid.Text('EDITED')),
                urwid.LineBox(
                    urwid.Columns([
                        (1, urwid.Text('')),
                        urwid.Text(content),
                    ]), tline='', bline='', rline='',
                    lline=MESSAGE_CONTENT_MARKER
                )
            ]),
            align='left', left=left_padding,
            width=('relative', 100), min_width=10, right=5)

        # Reactions
        reactions = self.reactions_view(self.message['reactions'])

        # Footlinks.
        footlinks = self.footlinks_view(self.message_links)

        # Build parts together and return
        parts = [
            (recipient_header, recipient_header is not None),
            (content_header, any_differences),
            (content, True),
            (footlinks, footlinks is not None),
            (reactions, reactions != ''),
        ]
        return [part for part, condition in parts if condition]

    def transform_content(self) -> Tuple[None, Any]:
        soup = BeautifulSoup(self.message['content'], 'lxml')
        body = soup.find(name='body')
        if body and body.find(name='blockquote'):
            self.indent_quoted_content(soup, QUOTED_TEXT_MARKER)

        return (None, self.soup2markup(body))

    def indent_quoted_content(self, soup: Any, padding_char: str) -> None:
        """
        We indent quoted text by padding them.
        The extent of indentation depends on their level of quoting.
        For example:
        [Before Padding]               [After Padding]

        <blockquote>                    <blockquote>
        <blockquote>                    <blockquote>
        <p>Foo</p>                      <p>▒ ▒ </p><p>Foo</p>
        </blockquote>       --->        </blockquote>
        <p>Boo</p>                      <p>▒ </p><p>Boo</p>
        </blockquote>                   </blockquote>
        """
        pad_count = 1
        blockquote_list = soup.find_all('blockquote')
        self.bq_len = len(blockquote_list)
        for tag in blockquote_list:
            child_list = tag.findChildren(recursive=False)
            actual_padding = (padding_char + ' ') * pad_count
            if len(child_list) == 1:
                pad_count = 0
                child_iterator = child_list
            else:
                child_iterator = child_list[1:]
            for child in child_iterator:
                new_tag = soup.new_tag('p')
                new_tag.string = actual_padding
                # If the quoted message is multi-line message
                # we deconstruct it and pad it at break-points (<br/>)
                if child.findAll('br'):
                    for br in child.findAll('br'):
                        next_s = br.nextSibling
                        text = str(next_s).strip()
                        if text:
                            insert_tag = soup.new_tag('p')
                            insert_tag.string = '\n' + actual_padding + text
                            next_s.replace_with(insert_tag)
                child.insert_before(new_tag)
            pad_count += 1

    def selectable(self) -> bool:
        # Returning True, indicates that this widget
        # is designed to take focus.
        return True

    def mouse_event(self, size: urwid_Size, event: str, button: int,
                    col: int, row: int, focus: bool) -> bool:
        if event == 'mouse press':
            if button == 1:
                if self.model.controller.is_in_editor_mode():
                    return True
                self.keypress(size, primary_key_for_command('ENTER'))
                return True
        elif event == 'mouse drag':
            selection_key = "Fn + Alt" if platform == "darwin" else "Shift"
            self.model.controller.view.set_footer_text([
                'Try pressing ',
                ('code', ' ' + selection_key + ' '),
                ' and dragging to select text.'
            ])
            self.displaying_selection_hint = True
        elif event == 'mouse release' and self.displaying_selection_hint:
            self.model.controller.view.set_footer_text()
            self.displaying_selection_hint = False

        return super().mouse_event(size, event, button, col, row, focus)

    def keypress(self, size: urwid_Size, key: str) -> Optional[str]:
        if is_command_key('ENTER', key):
            if self.message['type'] == 'private':
                self.model.controller.view.write_box.private_box_view(
                    email=self.recipients_emails,
                    recipient_user_ids=self.recipient_ids,
                )
            elif self.message['type'] == 'stream':
                self.model.controller.view.write_box.stream_box_view(
                    caption=self.message['display_recipient'],
                    title=self.message['subject'],
                    stream_id=self.stream_id,
                )
        elif is_command_key('STREAM_MESSAGE', key):
            if self.message['type'] == 'private':
                self.model.controller.view.write_box.private_box_view(
                    email=self.recipients_emails,
                    recipient_user_ids=self.recipient_ids,
                )
            elif self.message['type'] == 'stream':
                self.model.controller.view.write_box.stream_box_view(
                    caption=self.message['display_recipient'],
                    stream_id=self.stream_id,
                )
        elif is_command_key('STREAM_NARROW', key):
            if self.message['type'] == 'private':
                self.model.controller.narrow_to_user(self)
            elif self.message['type'] == 'stream':
                self.model.controller.narrow_to_stream(self)
        elif is_command_key('TOGGLE_NARROW', key):
            self.model.unset_search_narrow()
            if self.message['type'] == 'private':
                if (
                    len(self.model.narrow) == 1
                    and self.model.narrow[0][0] == 'pm_with'
                   ):
                    self.model.controller.show_all_pm(self)
                else:
                    self.model.controller.narrow_to_user(self)
            elif self.message['type'] == 'stream':
                if len(self.model.narrow) > 1:  # in a topic
                    self.model.controller.narrow_to_stream(self)
                else:
                    self.model.controller.narrow_to_topic(self)
        elif is_command_key('TOPIC_NARROW', key):
            if self.message['type'] == 'private':
                self.model.controller.narrow_to_user(self)
            elif self.message['type'] == 'stream':
                self.model.controller.narrow_to_topic(self)
        elif is_command_key('ALL_MESSAGES', key):
            self.model.controller.show_all_messages(self)
        elif is_command_key('REPLY_AUTHOR', key):
            # All subscribers from recipient_ids are not needed here.
            self.model.controller.view.write_box.private_box_view(
                email=self.message['sender_email'],
                recipient_user_ids=[self.message['sender_id']],
            )
        elif is_command_key('MENTION_REPLY', key):
            self.keypress(size, 'enter')
            mention = '@**' + self.message['sender_full_name'] + '** '
            self.model.controller.view.write_box.msg_write_box.set_edit_text(
                mention)
            self.model.controller.view.write_box.msg_write_box.set_edit_pos(
                len(mention))
            self.model.controller.view.middle_column.set_focus('footer')
        elif is_command_key('QUOTE_REPLY', key):
            self.keypress(size, 'enter')

            # To correctly quote a message that contains quote/code-blocks,
            # we need to fence quoted message containing ``` with ````,
            # ```` with ````` and so on.
            message_raw_content = self.model.client.get_raw_message(
                self.message['id'])['raw_content']
            fence = get_unused_fence(message_raw_content)

            absolute_url = near_message_url(
                self.model.server_url[:-1], self.message)

            # Compose box should look something like this:
            #   @_**Zeeshan|514** [said](link to message):
            #   ```quote
            #   message_content
            #   ```
            quote = '@_**{0}|{1}** [said]({2}):\n{3}quote\n{4}\n{3}\n'.format(
                        self.message['sender_full_name'],
                        self.message['sender_id'],
                        absolute_url,
                        fence,
                        message_raw_content)

            self.model.controller.view.write_box.msg_write_box.set_edit_text(
                quote)
            self.model.controller.view.write_box.msg_write_box.set_edit_pos(
                len(quote))
            self.model.controller.view.middle_column.set_focus('footer')
        elif is_command_key('EDIT_MESSAGE', key):
            if self.message['sender_id'] != self.model.user_id:
                self.model.controller.view.set_footer_text(
                        " You can't edit messages sent by other users.", 3)
                return key
            # Check if editing is allowed in the realm
            elif not self.model.initial_data['realm_allow_message_editing']:
                self.model.controller.view.set_footer_text(
                    " Editing sent message is disabled.", 3)
                return key
            # Check if message is still editable, i.e. within
            # the time limit.
            time_since_msg_sent = time() - self.message['timestamp']
            edit_time_limit = self.model.initial_data[
                    'realm_message_content_edit_limit_seconds']
            msg_body_edit_enabled = True
            if time_since_msg_sent >= edit_time_limit:
                if self.message['type'] == 'private':
                    self.model.controller.view.set_footer_text(
                            " Time Limit for editing the message has"
                            " been exceeded.", 3)
                    return key
                elif self.message['type'] == 'stream':
                    self.model.controller.view.set_footer_text(
                            " Only topic editing allowed."
                            " Time Limit for editing the message body has"
                            " been exceeded.", 3)
                    msg_body_edit_enabled = False

            if self.message['type'] == 'private':
                self.keypress(size, 'enter')
            elif self.message['type'] == 'stream':
                self.model.controller.view.write_box.stream_box_edit_view(
                    stream_id=self.stream_id,
                    caption=self.message['display_recipient'],
                    title=self.message['subject']
                )
            msg_id = self.message['id']
            msg = self.model.client.get_raw_message(msg_id)['raw_content']
            write_box = self.model.controller.view.write_box
            write_box.msg_edit_id = msg_id
            write_box.msg_write_box.set_edit_text(msg)
            write_box.msg_write_box.set_edit_pos(len(msg))
            write_box.msg_body_edit_enabled = msg_body_edit_enabled
            # Set focus to topic box if message body editing is disabled.
            if not msg_body_edit_enabled:
                write_box.focus_position = write_box.FOCUS_CONTAINER_HEADER
                (write_box.header_write_box.
                 focus_col) = write_box.FOCUS_HEADER_BOX_TOPIC

            self.model.controller.view.middle_column.set_focus('footer')
        elif is_command_key('MSG_INFO', key):
            self.model.controller.show_msg_info(self.message,
                                                self.message_links,
                                                self.time_mentions)
        return key


class SearchBox(urwid.Pile):
    def __init__(self, controller: Any) -> None:
        self.controller = controller
        super().__init__(self.main_view())

    def main_view(self) -> Any:
        search_text = ("Search ["
                       + ", ".join(keys_for_command("SEARCH_MESSAGES"))
                       + "]: ")
        self.text_box = ReadlineEdit(search_text + " ")
        # Add some text so that when packing,
        # urwid doesn't hide the widget.
        self.conversation_focus = urwid.Text(" ")
        self.search_bar = urwid.Columns([
            ('pack', self.conversation_focus),
            ('pack', urwid.Text("  ")),
            self.text_box,
        ])
        self.msg_narrow = urwid.Text("DONT HIDE")
        self.recipient_bar = urwid.LineBox(
            self.msg_narrow, title="Current message recipients",
            tline='─', lline='', trcorner='─', tlcorner='─',
            blcorner='─', rline='', bline='─', brcorner='─')
        return [self.search_bar, self.recipient_bar]

    def keypress(self, size: urwid_Size, key: str) -> Optional[str]:
        if ((is_command_key('ENTER', key) and self.text_box.edit_text == '')
                or is_command_key('GO_BACK', key)):
            self.text_box.set_edit_text("")
            self.controller.exit_editor_mode()
            self.controller.view.middle_column.set_focus('body')
            return key

        elif is_command_key('ENTER', key):
            self.controller.exit_editor_mode()
            self.controller.search_messages(self.text_box.edit_text)
            self.controller.view.middle_column.set_focus('body')
            return key

        key = super().keypress(size, key)
        return key


class PanelSearchBox(urwid.Edit):
    """
    Search Box to search panel views in real-time.
    """

    def __init__(self, panel_view: Any, search_command: str,
                 update_function: Callable[..., None]) -> None:
        self.panel_view = panel_view
        self.search_command = search_command
        self.search_text = ("Search ["
                            + ", ".join(keys_for_command(search_command))
                            + "]: ")
        urwid.connect_signal(self, 'change', update_function)
        super().__init__(caption='', edit_text=self.search_text)

    def reset_search_text(self) -> None:
        self.set_caption('')
        self.set_edit_text(self.search_text)

    def valid_char(self, ch: str) -> bool:
        # This method 'strips' leading space *before* entering it in the box
        if self.edit_text:
            # Use regular validation if already have text
            return super().valid_char(ch)
        elif len(ch) != 1:
            # urwid expands some unicode to strings to be useful
            # (so we need to work around eg 'backspace')
            return False
        else:
            # Skip unicode 'Control characters' and 'space Zeperators'
            # This includes various invalid characters and complex spaces
            return unicodedata.category(ch) not in ('Cc', 'Zs')

    def keypress(self, size: urwid_Size, key: str) -> Optional[str]:
        if ((is_command_key('ENTER', key) and self.get_edit_text() == '')
                or is_command_key('GO_BACK', key)):
            self.panel_view.view.controller.exit_editor_mode()
            self.reset_search_text()
            self.panel_view.set_focus("body")
            self.panel_view.keypress(size, 'esc')
        elif is_command_key('ENTER', key):
            self.panel_view.view.controller.exit_editor_mode()
            self.set_caption([('filter_results', 'Search Results'), ' '])
            self.panel_view.set_focus("body")
            if hasattr(self.panel_view, 'log') and len(self.panel_view.log):
                self.panel_view.body.set_focus(0)
        return super().keypress(size, key)



from typing import Any, List, Optional, Tuple, Union, cast


def parse_html_table(table_element: Any) -> Tuple[List[str], List[List[str]]]:
    """
    Parses an HTML table to extract cell items and column alignments.

    The table cells are stored in `cells` in a row-wise manner.
    cells = [[row0, row0, row0],
             [row1, row1, row1],
             [row2, row2, row2]]
    """
    headers = table_element.thead.tr.find_all('th')
    rows = table_element.tbody.find_all('tr')
    column_alignments = []

    # Add +1 to count the header row as well.
    cells = [[] for _ in range(len(rows) + 1)]  # type: List[List[str]]

    # Fill up `cells` with the header/0th row and extract alignments.
    for header in headers:
        cells[0].append(header.text)
        column_alignments.append(header.get(('align'), 'left'))

    # Fill up `cells` with body rows.
    for index, row in enumerate(rows, start=1):
        for tdata in row.find_all('td'):
            cells[index].append(tdata.text)
    return (column_alignments, cells)


StyledTableData = List[Union[str, Tuple[Optional[str], str]]]


def pad_row_strip(row_strip: StyledTableData, fill_char: str=' ',
                  fill_width: int=1) -> StyledTableData:
    """
    Returns back a padded row strip.

    This only pads the box-drawing unicode characters. In particular, all the
    connector characters are padded both sides, the leftmost character is
    padded right and the rightmost is padded left.

    The structure of `row_strip` for a table with three columns:
    row_strip = [
        leftmost_char,
        cell_content,
        connector_char,
        cell_content,
        connector_char,
        cell_content,
        rightmost_char,
    ]

    Note: `cast` is used for assisting mypy.
    """
    fill = fill_char * fill_width

    # Pad the leftmost box-drawing character.
    row_strip[0] = cast(str, row_strip[0]) + fill

    # Pad the connector box-drawing characters.
    for index in range(2, len(row_strip) - 1, 2):
        row_strip[index] = fill + cast(str, row_strip[index]) + fill

    # Pad the rightmost box-drawing character.
    row_strip[-1] = fill + cast(str, row_strip[-1])
    return row_strip


def row_with_styled_content(row: List[str], column_alignments: List[str],
                            column_widths: List[int], vertical_bar: str,
                            row_style: Optional[str]=None) -> StyledTableData:
    """
    Constructs styled row strip, for markup table, using unicode characters
    and row elements.
    """
    aligner = {'center': str.center, 'left': str.ljust, 'right': str.rjust}
    row_strip = [vertical_bar, ]  # type: StyledTableData
    for column_num, cell in enumerate(row):
        aligned_text = aligner[column_alignments[column_num]](
            cell, column_widths[column_num]
        )
        row_strip.extend([(row_style, aligned_text), vertical_bar])
    row_strip.pop()  # Remove the extra vertical_bar.
    row_strip.append(vertical_bar + '\n')
    return pad_row_strip(row_strip)


def row_with_only_border(lcorner: str, line: str, connector: str, rcorner: str,
                         column_widths: List[int],
                         newline: bool=True) -> StyledTableData:
    """
    Given left corner, line, connecter and right corner unicode character,
    constructs a border row strip for markup table.
    """
    border = [lcorner, ]  # type: StyledTableData
    for width in column_widths:
        border.extend([line * width, connector])
    border.pop()  # Remove the extra connector.
    if newline:
        rcorner += '\n'
    border.append(rcorner)
    return pad_row_strip(border, fill_char=line)


def render_table(table_element: Any) -> StyledTableData:
    """
    A helper function for rendering a markup table in the MessageBox.
    """
    column_alignments, cells = parse_html_table(table_element)

    # Calculate the width required for each column.
    column_widths = [
        len(max(column, key=lambda string: len(string)))
        for column in zip(*cells)
    ]

    top_border = row_with_only_border('┌', '─', '┬', '┐', column_widths)
    middle_border = row_with_only_border('├', '─', '┼', '┤', column_widths)
    bottom_border = row_with_only_border('└', '─', '┴', '┘', column_widths,
                                         newline=False)

    # Construct the table, row-by-row.
    table = []  # type: StyledTableData

    # Add the header/0th row and the borders that surround it to the table.
    table.extend(top_border)
    table.extend(row_with_styled_content(cells.pop(0), column_alignments,
                                         column_widths, '│',
                                         row_style='table_head'))
    table.extend(middle_border)

    # Add the body rows to the table followed by the bottom-most border in the
    # end.
    for row in cells:
        table.extend(row_with_styled_content(row, column_alignments,
                                             column_widths, '│'))
    table.extend(bottom_border)

    return table



import threading
import time
from collections import OrderedDict
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

import urwid
from typing_extensions import Literal

from zulipterminal.config.keys import (
    HELP_CATEGORIES,
    KEY_BINDINGS,
    is_command_key,
    keys_for_command,
)
from zulipterminal.config.symbols import (
    CHECK_MARK,
    LIST_TITLE_BAR_LINE,
    PINNED_STREAMS_DIVIDER,
    STATUS_ACTIVE,
    STATUS_IDLE,
    STATUS_INACTIVE,
    STATUS_OFFLINE,
    STREAM_MARKER_PRIVATE,
    STREAM_MARKER_PUBLIC,
)
from zulipterminal.helper import (
    Message,
    asynch,
    edit_mode_captions,
    match_stream,
    match_user,
)
from zulipterminal.ui_tools.boxes import PanelSearchBox
from zulipterminal.ui_tools.buttons import (
    HomeButton,
    MentionedButton,
    MessageLinkButton,
    PMButton,
    StarredButton,
    StreamButton,
    TopicButton,
    UnreadPMButton,
    UserButton,
)
from zulipterminal.ui_tools.utils import create_msg_box_list
from zulipterminal.urwid_types import urwid_Size


class ModListWalker(urwid.SimpleFocusListWalker):

    def set_focus(self, position: int) -> None:
        # When setting focus via set_focus method.
        self.focus = position
        self._modified()
        if hasattr(self, 'read_message'):
            self.read_message()

    def _set_focus(self, index: int) -> None:
        # This method is called when directly setting focus via
        # self.focus = focus_position
        if not self:
            self._focus = 0
            return
        if index < 0 or index >= len(self):
            raise IndexError('focus index is out of range: %s' % (index,))
        if index != int(index):
            raise IndexError('invalid focus index: %s' % (index,))
        index = int(index)
        if index != self._focus:
            self._focus_changed(index)
        self._focus = index
        if hasattr(self, 'read_message'):
            self.read_message()

    def extend(self, items: List[Any],
               focus_position: Optional[int]=None) -> int:
        if focus_position is None:
            focus = self._adjust_focus_on_contents_modified(
                slice(len(self), len(self)), items)
        else:
            focus = focus_position
        rval = super(urwid.MonitoredFocusList, self).extend(items)
        self._set_focus(focus)
        return rval


class MessageView(urwid.ListBox):
    def __init__(self, model: Any, view: Any) -> None:
        self.model = model
        self.view = view
        # Initialize for reference
        self.focus_msg = 0
        self.log = ModListWalker(self.main_view())
        self.log.read_message = self.read_message

        super().__init__(self.log)
        self.set_focus(self.focus_msg)
        # if loading new/old messages - True
        self.old_loading = False
        self.new_loading = False

    def main_view(self) -> List[Any]:
        msg_btn_list = create_msg_box_list(self.model)
        focus_msg = self.model.get_focus_in_current_narrow()
        if focus_msg == set():
            focus_msg = len(msg_btn_list) - 1
        self.focus_msg = focus_msg
        return msg_btn_list

    @asynch
    def load_old_messages(self, anchor: int=10000000000) -> None:
        self.old_loading = True

        ids_to_keep = self.model.get_message_ids_in_current_narrow()
        if self.log:
            top_message_id = self.log[0].original_widget.message['id']
            ids_to_keep.remove(top_message_id)  # update this id
            no_update_baseline = {top_message_id}
        else:
            no_update_baseline = set()

        self.model.get_messages(num_before=30, num_after=0, anchor=anchor)
        ids_to_process = (self.model.get_message_ids_in_current_narrow()
                          - ids_to_keep)

        # Only update if more messages are provided
        if ids_to_process != no_update_baseline:
            if self.log:
                self.log.remove(self.log[0])  # avoid duplication when updating

            message_list = create_msg_box_list(self.model, ids_to_process)
            message_list.reverse()
            for msg_w in message_list:
                self.log.insert(0, msg_w)

            self.set_focus(self.focus_msg)  # Return focus to original message

            self.model.controller.update_screen()

        self.old_loading = False

    @asynch
    def load_new_messages(self, anchor: int) -> None:
        self.new_loading = True
        current_ids = self.model.get_message_ids_in_current_narrow()
        self.model.get_messages(num_before=0, num_after=30, anchor=anchor)
        new_ids = self.model.get_message_ids_in_current_narrow() - current_ids
        if self.log:
            last_message = self.log[-1].original_widget.message
        else:
            last_message = None

        message_list = create_msg_box_list(self.model, new_ids,
                                           last_message=last_message)
        self.log.extend(message_list)

        self.model.controller.update_screen()
        self.new_loading = False

    def mouse_event(self, size: urwid_Size, event: str, button: int, col: int,
                    row: int, focus: bool) -> bool:
        if event == 'mouse press':
            if button == 4:
                self.keypress(size, 'up')
                return True
            if button == 5:
                self.keypress(size, 'down')
                return True
        return super().mouse_event(size, event, button, col, row, focus)

    def keypress(self, size: urwid_Size, key: str) -> Optional[str]:
        if is_command_key('GO_DOWN', key) and not self.new_loading:
            try:
                position = self.log.next_position(self.focus_position)
                self.set_focus(position, 'above')
                self.set_focus_valign('middle')

                return key
            except Exception:
                if self.focus:
                    id = self.focus.original_widget.message['id']
                    self.load_new_messages(id)
                return key

        elif is_command_key('GO_UP', key) and not self.old_loading:
            try:
                position = self.log.prev_position(self.focus_position)
                self.set_focus(position, 'below')
                self.set_focus_valign('middle')
                return key
            except Exception:
                if self.focus:
                    id = self.focus.original_widget.message['id']
                    self.load_old_messages(id)
                else:
                    self.load_old_messages()
                return key

        elif is_command_key('SCROLL_UP', key) and not self.old_loading:
            if (self.focus is not None
                    and self.focus_position == 0):
                return self.keypress(size, 'up')
            else:
                return super().keypress(size, 'page up')

        elif is_command_key('SCROLL_DOWN', key) and not self.old_loading:
            if (self.focus is not None
                    and self.focus_position == len(self.log) - 1):
                return self.keypress(size, 'down')
            else:
                return super().keypress(size, 'page down')

        elif is_command_key('THUMBS_UP', key):
            if self.focus is not None:
                self.model.react_to_message(
                    self.focus.original_widget.message,
                    reaction_to_toggle='thumbs_up')

        elif is_command_key('TOGGLE_STAR_STATUS', key):
            if self.focus is not None:
                message = self.focus.original_widget.message
                self.model.toggle_message_star_status(message)

        key = super().keypress(size, key)
        return key

    def update_search_box_narrow(self, message_view: Any) -> None:
        if not hasattr(self.model.controller, 'view'):
            return
        # if view is ready display current narrow
        # at the bottom of the view.
        recipient_bar = message_view.top_header_bar(message_view)
        top_header = message_view.top_search_bar()
        self.model.controller.view.search_box.conversation_focus.set_text(
            top_header.markup
            )
        self.model.controller.view.search_box.msg_narrow.set_text(
            recipient_bar.markup
            )
        self.model.controller.update_screen()

    def read_message(self, index: int=-1) -> None:
        # Message currently in focus
        if hasattr(self.model.controller, "view"):
            view = self.model.controller.view
        else:
            return
        msg_w, curr_pos = self.body.get_focus()
        if msg_w is None:
            return
        self.update_search_box_narrow(msg_w.original_widget)

        # Do not read messages in explore mode.
        if self.model.controller.in_explore_mode:
            return

        # Do not read messages in any search narrow.
        if self.model.is_search_narrow():
            return

        # If this the last message in the view and focus is set on this message
        # then read the message.
        last_message_focused = (curr_pos == len(self.log) - 1)
        # Only allow reading a message when middle column is
        # in focus.
        if not(view.body.focus_col == 1 or last_message_focused):
            return
        # save the current focus
        self.model.set_focus_in_current_narrow(self.focus_position)
        # msg ids that have been read
        read_msg_ids = list()
        # until we find a read message above the current message
        while msg_w.attr_map == {None: 'unread'}:
            msg_id = msg_w.original_widget.message['id']
            read_msg_ids.append(msg_id)
            self.model.index['messages'][msg_id]['flags'].append('read')
            msg_w.set_attr_map({None: None})
            msg_w, curr_pos = self.body.get_prev(curr_pos)
            if msg_w is None:
                break
        self.model.mark_message_ids_as_read(read_msg_ids)


class StreamsViewDivider(urwid.Divider):
    """
    A custom urwid.Divider to visually separate pinned and unpinned streams.
    """

    def __init__(self) -> None:
        # FIXME: Necessary since the divider is treated as a StreamButton.
        # NOTE: This is specifically for stream search to work correctly.
        self.stream_id = -1
        self.stream_name = ''
        super().__init__(div_char=PINNED_STREAMS_DIVIDER)


class StreamsView(urwid.Frame):
    def __init__(self, streams_btn_list: List[Any], view: Any) -> None:
        self.view = view
        self.log = urwid.SimpleFocusListWalker(streams_btn_list)
        self.streams_btn_list = streams_btn_list
        self.focus_index_before_search = 0
        list_box = urwid.ListBox(self.log)
        self.stream_search_box = PanelSearchBox(self,
                                                'SEARCH_STREAMS',
                                                self.update_streams)
        super().__init__(list_box, header=urwid.LineBox(
            self.stream_search_box, tlcorner='─', tline='', lline='',
            trcorner='─', blcorner='─', rline='',
            bline='─', brcorner='─'
        ))
        self.search_lock = threading.Lock()

    @asynch
    def update_streams(self, search_box: Any, new_text: str) -> None:
        if not self.view.controller.is_in_editor_mode():
            return
        # wait for any previously started search to finish to avoid
        # displaying wrong stream list.
        with self.search_lock:
            stream_buttons = [
                (stream, stream.stream_name)
                for stream in self.streams_btn_list.copy()
            ]
            streams_display = match_stream(stream_buttons, new_text,
                                           self.view.pinned_streams)[0]

            # Add a divider to separate pinned streams from the rest.
            pinned_stream_names = [
                stream['name']
                for stream in self.view.pinned_streams
            ]
            first_unpinned_index = len(streams_display)
            for index, stream in enumerate(streams_display):
                if stream.stream_name not in pinned_stream_names:
                    first_unpinned_index = index
                    break
            if first_unpinned_index not in [0, len(streams_display)]:
                # Do not add a divider when it is already present. This can
                # happen when new_text=''.
                if not isinstance(streams_display[first_unpinned_index],
                                  StreamsViewDivider):
                    streams_display.insert(first_unpinned_index,
                                           StreamsViewDivider())

            self.log.clear()
            self.log.extend(streams_display)
            self.view.controller.update_screen()

    def mouse_event(self, size: urwid_Size, event: str, button: int, col: int,
                    row: int, focus: bool) -> bool:
        if event == 'mouse press':
            if button == 4:
                self.keypress(size, 'up')
                return True
            elif button == 5:
                self.keypress(size, 'down')
                return True
        return super().mouse_event(size, event, button, col, row, focus)

    def keypress(self, size: urwid_Size, key: str) -> Optional[str]:
        if is_command_key('SEARCH_STREAMS', key):
            self.set_focus('header')
            return key
        elif is_command_key('GO_BACK', key):
            self.stream_search_box.reset_search_text()
            self.log.clear()
            self.log.extend(self.streams_btn_list)
            self.set_focus('body')
            self.log.set_focus(self.focus_index_before_search)
            self.view.controller.update_screen()
            return key
        return_value = super().keypress(size, key)
        _, self.focus_index_before_search = self.log.get_focus()
        return return_value


class TopicsView(urwid.Frame):
    def __init__(self, topics_btn_list: List[Any], view: Any,
                 stream_button: Any) -> None:
        self.view = view
        self.log = urwid.SimpleFocusListWalker(topics_btn_list)
        self.topics_btn_list = topics_btn_list
        self.stream_button = stream_button
        self.focus_index_before_search = 0
        self.list_box = urwid.ListBox(self.log)
        self.topic_search_box = PanelSearchBox(self,
                                               'SEARCH_TOPICS',
                                               self.update_topics)
        self.header_list = urwid.Pile([self.stream_button,
                                       urwid.Divider('─'),
                                       self.topic_search_box])
        super().__init__(self.list_box, header=urwid.LineBox(
            self.header_list, tlcorner='─', tline='', lline='',
            trcorner='─', blcorner='─', rline='',
            bline='─', brcorner='─'
        ))
        self.search_lock = threading.Lock()

    @asynch
    def update_topics(self, search_box: Any, new_text: str) -> None:
        if not self.view.controller.is_in_editor_mode():
            return
        # wait for any previously started search to finish to avoid
        # displaying wrong topics list.
        with self.search_lock:
            new_text = new_text.lower()
            topics_to_display = [
                topic
                for topic in self.topics_btn_list.copy()
                if new_text in topic.topic_name.lower()
            ]
            self.log.clear()
            self.log.extend(topics_to_display)
            self.view.controller.update_screen()

    def update_topics_list(self, stream_id: int, topic_name: str,
                           sender_id: int) -> None:
        # More recent topics are found towards the beginning
        # of the list.
        for topic_iterator, topic_button in enumerate(self.log):
            if topic_button.topic_name == topic_name:
                self.log.insert(0, self.log.pop(topic_iterator))
                self.list_box.set_focus_valign('bottom')
                if sender_id == self.view.model.user_id:
                    self.list_box.set_focus(0)
                return
        # No previous topics with same topic names are found
        # hence we create a new topic button for it.
        new_topic_button = TopicButton(stream_id,
                                       topic_name,
                                       self.view.controller,
                                       self.view.LEFT_WIDTH,
                                       0)
        self.log.insert(0, new_topic_button)
        self.list_box.set_focus_valign('bottom')
        if sender_id == self.view.model.user_id:
            self.list_box.set_focus(0)

    def mouse_event(self, size: urwid_Size, event: str, button: int, col: int,
                    row: int, focus: bool) -> bool:
        if event == 'mouse press':
            if button == 4:
                self.keypress(size, 'up')
                return True
            elif button == 5:
                self.keypress(size, 'down')
                return True
        return super().mouse_event(size, event, button, col, row, focus)

    def keypress(self, size: urwid_Size, key: str) -> Optional[str]:
        if is_command_key('TOGGLE_TOPIC', key):
            # Exit topic view
            self.view.left_panel.show_stream_view()
        elif is_command_key('GO_RIGHT', key):
            self.view.show_left_panel(visible=False)
            self.view.body.focus_col = 1
        if is_command_key('SEARCH_TOPICS', key):
            self.set_focus('header')
            self.header_list.set_focus(2)
            return key
        elif is_command_key('GO_BACK', key):
            self.topic_search_box.reset_search_text()
            self.log.clear()
            self.log.extend(self.topics_btn_list)
            self.set_focus('body')
            self.log.set_focus(self.focus_index_before_search)
            self.view.controller.update_screen()
            return key
        return_value = super().keypress(size, key)
        _, self.focus_index_before_search = self.log.get_focus()
        return return_value


class UsersView(urwid.ListBox):
    def __init__(self, controller: Any, users_btn_list: List[Any]) -> None:
        self.log = urwid.SimpleFocusListWalker(users_btn_list)
        self.controller = controller
        super().__init__(self.log)

    def mouse_event(self, size: urwid_Size, event: str, button: int, col: int,
                    row: int, focus: bool) -> bool:
        if event == 'mouse press':
            if button == 1:
                if self.controller.is_in_editor_mode():
                    return True
            if button == 4:
                for _ in range(5):
                    self.keypress(size, 'up')
                return True
            elif button == 5:
                for _ in range(5):
                    self.keypress(size, 'down')
        return super().mouse_event(size, event, button, col, row, focus)


class MiddleColumnView(urwid.Frame):
    def __init__(self, view: Any, model: Any,
                 write_box: Any, search_box: Any) -> None:
        message_view = MessageView(model, view)
        self.model = model
        self.controller = model.controller
        self.view = view
        self.last_unread_topic = None
        self.last_unread_pm = None
        self.search_box = search_box
        view.message_view = message_view
        super().__init__(message_view, header=search_box, footer=write_box)

    def get_next_unread_topic(self) -> Optional[Tuple[int, str]]:
        topics = list(self.model.unread_counts['unread_topics'].keys())
        next_topic = False
        for topic in topics:
            if next_topic is True:
                self.last_unread_topic = topic
                return topic
            if topic == self.last_unread_topic:
                next_topic = True
        if len(topics) > 0:
            topic = topics[0]
            self.last_unread_topic = topic
            return topic
        return None

    def get_next_unread_pm(self) -> Optional[int]:
        pms = list(self.model.unread_counts['unread_pms'].keys())
        next_pm = False
        for pm in pms:
            if next_pm is True:
                self.last_unread_pm = pm
                return pm
            if pm == self.last_unread_pm:
                next_pm = True
        if len(pms) > 0:
            pm = pms[0]
            self.last_unread_pm = pm
            return pm
        return None

    def keypress(self, size: urwid_Size, key: str) -> Optional[str]:
        if is_command_key('GO_BACK', key):
            self.header.keypress(size, 'esc')
            self.footer.keypress(size, 'esc')
            self.set_focus('body')

        elif self.focus_position in ['footer', 'header']:
            return super().keypress(size, key)

        elif is_command_key('SEARCH_MESSAGES', key):
            self.controller.enter_editor_mode_with(self.search_box)
            self.set_focus('header')
            return key

        elif is_command_key('REPLY_MESSAGE', key):
            self.body.keypress(size, 'enter')
            if self.footer.focus is not None:
                self.set_focus('footer')
                self.footer.focus_position = 1
            return key

        elif is_command_key('STREAM_MESSAGE', key):
            self.body.keypress(size, 'c')
            # For new streams with no previous conversation.
            if self.footer.focus is None:
                stream_id = self.model.stream_id
                stream_dict = self.model.stream_dict
                self.footer.stream_box_view(
                    caption=stream_dict[stream_id]['name'])
            self.set_focus('footer')
            self.footer.focus_position = 0
            return key

        elif is_command_key('REPLY_AUTHOR', key):
            self.body.keypress(size, 'R')
            if self.footer.focus is not None:
                self.set_focus('footer')
                self.footer.focus_position = 1
            return key

        elif is_command_key('NEXT_UNREAD_TOPIC', key):
            # narrow to next unread topic
            stream_topic = self.get_next_unread_topic()
            if stream_topic is None:
                return key
            stream, topic = stream_topic
            self.controller.narrow_to_topic(TopicButton(stream, topic,
                                                        self.controller))
            return key
        elif is_command_key('NEXT_UNREAD_PM', key):
            # narrow to next unread pm
            pm = self.get_next_unread_pm()
            if pm is None:
                return key
            email = self.model.user_id_email_dict[pm]
            self.controller.narrow_to_user(UnreadPMButton(pm, email))
        elif is_command_key('PRIVATE_MESSAGE', key):
            # Create new PM message
            self.footer.private_box_view()
            self.set_focus('footer')
            self.footer.focus_position = 0
            return key
        elif is_command_key('GO_LEFT', key):
            self.view.show_left_panel(visible=True)
        elif is_command_key('GO_RIGHT', key):
            self.view.show_right_panel(visible=True)
        return super().keypress(size, key)


class RightColumnView(urwid.Frame):
    """
    Displays the users list on the right side of the app.
    """

    def __init__(self, width: int, view: Any) -> None:
        self.width = width
        self.view = view
        self.user_search = PanelSearchBox(self,
                                          'SEARCH_PEOPLE',
                                          self.update_user_list)
        self.view.user_search = self.user_search
        search_box = urwid.LineBox(
            self.user_search, tlcorner='─', tline='', lline='',
            trcorner='─', blcorner='─', rline='',
            bline='─', brcorner='─'
            )
        self.allow_update_user_list = True
        self.search_lock = threading.Lock()
        super().__init__(self.users_view(), header=search_box)

    @asynch
    def update_user_list(self, search_box: Any=None,
                         new_text: Optional[str]=None,
                         user_list: Any=None) -> None:
        """
        Updates user list via PanelSearchBox and _start_presence_updates.
        """
        assert (
            (user_list is None and search_box is not None)  # PanelSearchBox.
            or (user_list is not None and search_box is None
                and new_text is None)  # _start_presence_updates.
        )

        # Return if the method is called by PanelSearchBox (urwid.Edit) while
        # the search is inactive and user_list is None.
        # NOTE: The additional not user_list check is to not false trap
        # _start_presence_updates but allow it to update the user list.
        if not self.view.controller.is_in_editor_mode() and not user_list:
            return

        # Return if the method is called from _start_presence_updates while the
        # search, via PanelSearchBox, is active.
        if not self.allow_update_user_list and new_text is None:
            return

        # wait for any previously started search to finish to avoid
        # displaying wrong user list.
        with self.search_lock:
            if user_list:
                self.view.users = user_list
            users = self.view.users.copy()
            if new_text:
                users_display = [
                    user for user in users if match_user(user, new_text)
                ]
            else:
                users_display = users
            self.body = self.users_view(users_display)
            self.set_body(self.body)
            self.view.controller.update_screen()

    def users_view(self, users: Any=None) -> Any:
        reset_default_view_users = False
        if users is None:
            users = self.view.users.copy()
            reset_default_view_users = True

        state_icon = {
            'active': STATUS_ACTIVE,
            'idle': STATUS_IDLE,
            'offline': STATUS_OFFLINE,
            'inactive': STATUS_INACTIVE,
        }

        users_btn_list = list()
        for user in users:
            status = user['status']
            # Only include `inactive` users in search result.
            if (status == 'inactive'
                    and not self.view.controller.is_in_editor_mode()):
                continue
            unread_count = (self.view.model.unread_counts['unread_pms'].
                            get(user['user_id'], 0))
            is_current_user = (user['user_id'] == self.view.model.user_id)
            users_btn_list.append(
                UserButton(
                    user,
                    controller=self.view.controller,
                    view=self.view,
                    width=self.width,
                    state_marker=state_icon[status],
                    color='user_' + status,
                    count=unread_count,
                    is_current_user=is_current_user
                )
            )
        user_w = UsersView(self.view.controller, users_btn_list)
        # Donot reset them while searching.
        if reset_default_view_users:
            self.users_btn_list = users_btn_list
            self.view.user_w = user_w
        return user_w

    def keypress(self, size: urwid_Size, key: str) -> Optional[str]:
        if is_command_key('SEARCH_PEOPLE', key):
            self.allow_update_user_list = False
            self.set_focus('header')
            return key
        elif is_command_key('GO_BACK', key):
            self.user_search.reset_search_text()
            self.allow_update_user_list = True
            self.body = UsersView(self.view.controller,
                                  self.users_btn_list)
            self.set_body(self.body)
            self.set_focus('body')
            self.view.controller.update_screen()
            return key
        elif is_command_key('GO_LEFT', key):
            self.view.show_right_panel(visible=False)
        return super().keypress(size, key)


class LeftColumnView(urwid.Pile):
    """
    Displays the buttons at the left column of the app.
    """

    def __init__(self, width: int, view: Any) -> None:
        self.model = view.model
        self.view = view
        self.controller = view.controller
        self.width = width
        self.menu_v = self.menu_view()
        self.stream_v = self.streams_view()

        self.is_in_topic_view = False
        contents = [
            (4, self.menu_v),
            self.stream_v
        ]
        super().__init__(contents)

    def menu_view(self) -> Any:
        count = self.model.unread_counts.get('all_msg', 0)
        self.view.home_button = HomeButton(self.controller,
                                           count=count,
                                           width=self.width)

        count = self.model.unread_counts.get('all_pms', 0)
        self.view.pm_button = PMButton(self.controller,
                                       count=count,
                                       width=self.width)

        self.view.mentioned_button = MentionedButton(
            self.controller,
            width=self.width,
            count=self.model.unread_counts['all_mentions'])

        # Starred messages are by definition read already
        self.view.starred_button = StarredButton(self.controller,
                                                 width=self.width)

        menu_btn_list = [
            self.view.home_button,
            self.view.pm_button,
            self.view.mentioned_button,
            self.view.starred_button,
        ]
        w = urwid.ListBox(urwid.SimpleFocusListWalker(menu_btn_list))
        return w

    def streams_view(self) -> Any:
        streams_btn_list = [
                StreamButton(
                    stream,
                    controller=self.controller,
                    view=self.view,
                    width=self.width,
                    count=self.model.unread_counts['streams'].get(
                                    stream['id'], 0)
                ) for stream in self.view.pinned_streams]

        if len(streams_btn_list):
            streams_btn_list += [StreamsViewDivider()]

        streams_btn_list += [
                StreamButton(
                    stream,
                    controller=self.controller,
                    view=self.view,
                    width=self.width,
                    count=self.model.unread_counts['streams'].get(
                                    stream['id'], 0)
                ) for stream in self.view.unpinned_streams]

        self.view.stream_id_to_button = {stream.stream_id: stream
                                         for stream in streams_btn_list
                                         if hasattr(stream, 'stream_id')}

        self.view.stream_w = StreamsView(streams_btn_list, self.view)
        w = urwid.LineBox(
            self.view.stream_w, title="Streams",
            tlcorner=LIST_TITLE_BAR_LINE,
            tline=LIST_TITLE_BAR_LINE,
            trcorner=LIST_TITLE_BAR_LINE,
            blcorner='', rline='', lline='',
            bline='', brcorner='─'
            )
        return w

    def topics_view(self, stream_button: Any) -> Any:
        stream_id = stream_button.stream_id
        topics = self.model.topics_in_stream(stream_id)
        topics_btn_list = [
            TopicButton(
                stream_id=stream_id,
                topic=topic,
                controller=self.controller,
                width=self.width,
                count=self.model.unread_counts['unread_topics'].
                get((stream_id, topic), 0)
            )
            for topic in topics
        ]

        self.view.topic_w = TopicsView(topics_btn_list, self.view,
                                       stream_button)
        w = urwid.LineBox(
            self.view.topic_w, title="Topics",
            tlcorner=LIST_TITLE_BAR_LINE,
            tline=LIST_TITLE_BAR_LINE,
            trcorner=LIST_TITLE_BAR_LINE,
            blcorner='', rline='', lline='',
            bline='', brcorner='─'
            )
        return w

    def is_in_topic_view_with_stream_id(self, stream_id: int) -> bool:
        return (
            self.is_in_topic_view
            and stream_id == self.view.topic_w.stream_button.stream_id
        )

    def update_stream_view(self) -> None:
        self.stream_v = self.streams_view()
        if not self.is_in_topic_view:
            self.show_stream_view()

    def show_stream_view(self) -> None:
        self.is_in_topic_view = False
        self.contents[1] = (
            self.stream_v,
            self.options(height_type="weight")
        )

    def show_topic_view(self, stream_button: Any) -> None:
        self.is_in_topic_view = True
        self.contents[1] = (
            self.topics_view(stream_button),
            self.options(height_type="weight")
        )

    def keypress(self, size: urwid_Size, key: str) -> Optional[str]:
        if (
            is_command_key('SEARCH_STREAMS', key)
            or is_command_key('SEARCH_TOPICS', key)
           ):
            self.focus_position = 1
            if self.is_in_topic_view:
                self.view.topic_w.keypress(size, key)
            else:
                self.view.stream_w.keypress(size, key)
            return key
        elif is_command_key('GO_RIGHT', key):
            self.view.show_left_panel(visible=False)
        return super().keypress(size, key)


PopUpViewTableContent = Sequence[Tuple[str, Sequence[Union[str,
                                                           Tuple[str, str]]]]]


class PopUpView(urwid.ListBox):
    def __init__(self, controller: Any, widgets: List[Any],
                 command: str, requested_width: int, title: str) -> None:
        self.controller = controller
        self.command = command
        self.title = title
        self.log = urwid.SimpleFocusListWalker(widgets)

        max_cols, max_rows = controller.maximum_popup_dimensions()

        self.width = min(max_cols, requested_width)

        height = self.calculate_popup_height(widgets, self.width)
        self.height = min(max_rows, height)

        super().__init__(self.log)

    @staticmethod
    def calculate_popup_height(widgets: List[Any], popup_width: int) -> int:
        """
        Returns popup height. The popup height is calculated using urwid's
        .rows method on every widget.
        """
        return sum(widget.rows((popup_width, )) for widget in widgets)

    @staticmethod
    def calculate_table_widths(contents: PopUpViewTableContent,
                               title_len: int,
                               dividechars: int=2) -> Tuple[int, List[int]]:
        """
        Returns a tuple that contains the required width for the popup and a
        list that has column widths.
        """
        # Add 4 (for 2 Unicode characters on either side) to the popup title
        # length to make sure that the title gets displayed even when the
        # content or the category is shorter than the title length (+4 Unicode
        # characters).
        title_width = title_len + 4

        category_width = 0
        text_width = 0
        strip_widths = []
        for category, content in contents:
            category_width = max(category_width, len(category))
            for row in content:
                if isinstance(row, str):
                    # Measure the longest line if the text is separated by
                    # newline(s).
                    text_width = max(text_width, len(max(row.split('\n'),
                                                         key=len)))
                elif isinstance(row, tuple):
                    # Measure the longest line if the text is separated by
                    # newline(s).
                    max_row_lengths = [
                        len(max(text.split('\n'), key=len))
                        for text in row
                    ]
                    strip_widths.append(max_row_lengths)
        column_widths = [max(width) for width in zip(*strip_widths)]

        popup_width = max(sum(column_widths) + dividechars, title_width,
                          category_width, text_width)
        return (popup_width, column_widths)

    @staticmethod
    def make_table_with_categories(contents: PopUpViewTableContent,
                                   column_widths: List[int],
                                   dividechars: int=2) -> List[Any]:
        """
        Returns a list of widgets to render a table with different categories.
        """
        widgets = []  # type: List[Any]
        for category, content in contents:
            if category:
                if len(widgets) > 0:  # Separate categories with newline.
                    widgets.append(urwid.Text(''))
                widgets.append(urwid.Text(('popup_category', category)))
            for index, row in enumerate(content):
                if isinstance(row, str) and row:
                    widgets.append(urwid.Text(row))
                elif isinstance(row, tuple):
                    label, data = row
                    strip = urwid.Columns([
                            (column_widths[0], urwid.Text(label)),
                            urwid.Text(data)
                        ], dividechars=dividechars)
                    widgets.append(urwid.AttrWrap(
                        strip, None if index % 2 else 'popup_contrast')
                    )
        return widgets

    def keypress(self, size: urwid_Size, key: str) -> str:
        if is_command_key('GO_BACK', key) or is_command_key(self.command, key):
            self.controller.exit_popup()
        elif is_command_key('GO_UP', key):
            key = 'up'
        elif is_command_key('GO_DOWN', key):
            key = 'down'
        elif is_command_key('SCROLL_UP', key):
            key = 'page up'
        elif is_command_key('SCROLL_DOWN', key):
            key = 'page down'
        elif is_command_key('GO_TO_BOTTOM', key):
            key = 'end'
        return super().keypress(size, key)


class NoticeView(PopUpView):
    def __init__(self, controller: Any,
                 notice_text: str,
                 width: int,
                 title: str) -> None:
        widgets = [
            urwid.Divider(),
            urwid.Padding(urwid.Text(notice_text), left=1, right=1),
            urwid.Divider(),
        ]
        super().__init__(controller, widgets, 'GO_BACK', width, title)


class AboutView(PopUpView):
    def __init__(self, controller: Any, title: str, *, zt_version: str,
                 server_version: str,
                 server_feature_level: Optional[int],
                 theme_name: str, color_depth: int,
                 autohide_enabled: bool, footlink_enabled: bool) -> None:
        self.feature_level_content = (
            [('Feature level', str(server_feature_level))]
            if server_feature_level else []
        )
        contents = [
            ('Application', [('Zulip Terminal', zt_version)]),
            ('Server', [('Version', server_version)]
             + self.feature_level_content),
            ('Application Configuration', [
                ('Theme', theme_name),
                ('Autohide', 'enabled' if autohide_enabled else 'disabled'),
                ('Footlink', 'enabled' if footlink_enabled else 'disabled'),
                ('Color depth', str(color_depth))])
        ]

        popup_width, column_widths = self.calculate_table_widths(contents,
                                                                 len(title))
        widgets = self.make_table_with_categories(contents, column_widths)

        super().__init__(controller, widgets, 'ABOUT', popup_width, title)


class HelpView(PopUpView):
    def __init__(self, controller: Any, title: str) -> None:
        help_menu_content = []
        for category in HELP_CATEGORIES:
            keys_in_category = (binding for binding in KEY_BINDINGS.values()
                                if binding['key_category'] == category)
            key_bindings = []
            for binding in keys_in_category:
                key_bindings.append((binding['help_text'],
                                     ', '.join(binding['keys'])))
            help_menu_content.append((HELP_CATEGORIES[category], key_bindings))

        popup_width, column_widths = self.calculate_table_widths(
            help_menu_content, len(title))
        widgets = self.make_table_with_categories(help_menu_content,
                                                  column_widths)

        super().__init__(controller, widgets, 'HELP', popup_width, title)


class PopUpConfirmationView(urwid.Overlay):
    def __init__(self, controller: Any, question: Any,
                 success_callback: Callable[[], None]):
        self.controller = controller
        self.success_callback = success_callback
        yes = urwid.Button('Yes', self.exit_popup_yes)
        no = urwid.Button('No', self.exit_popup_no)
        yes._w = urwid.AttrMap(urwid.SelectableIcon(
            'Yes', 4), None, 'selected')
        no._w = urwid.AttrMap(urwid.SelectableIcon(
            'No', 4), None, 'selected')
        display_widget = urwid.GridFlow([yes, no], 3, 5, 1, 'center')
        wrapped_widget = urwid.WidgetWrap(display_widget)
        prompt = urwid.LineBox(
            urwid.ListBox(
                urwid.SimpleFocusListWalker(
                    [question, urwid.Divider(), wrapped_widget]
                )))
        urwid.Overlay.__init__(self, prompt, self.controller.view,
                               align="left", valign="top",
                               width=self.controller.view.LEFT_WIDTH + 1,
                               height=8)

    def exit_popup_yes(self, args: Any) -> None:
        self.success_callback()
        self.controller.exit_popup()

    def exit_popup_no(self, args: Any) -> None:
        self.controller.exit_popup()

    def keypress(self, size: urwid_Size, key: str) -> str:
        if is_command_key('GO_BACK', key):
            self.controller.exit_popup()
        return super().keypress(size, key)


class StreamInfoView(PopUpView):
    def __init__(self, controller: Any, stream_id: int) -> None:
        self.stream_id = stream_id
        self.controller = controller
        stream = controller.model.stream_dict[stream_id]

        total_members = len(stream['subscribers'])
        member_keys = ', '.join(map(repr, keys_for_command('STREAM_MEMBERS')))

        weekly_msg_count = stream['stream_weekly_traffic']

        stream_marker = (STREAM_MARKER_PRIVATE if stream['invite_only']
                         else STREAM_MARKER_PUBLIC)
        title = '{} {}'.format(stream_marker, stream['name'])

        desc = stream['description']

        stream_info_content = [('', [desc]),
                               ('Stream Details', [
                                   ('Weekly Message Count',
                                    str(weekly_msg_count)),
                                   ('Stream Members',
                                    '{} (Press {} to view list)'.format(
                                        total_members, member_keys))
                               ]),
                               ('Stream settings', [])]
        popup_width, column_widths = self.calculate_table_widths(
            stream_info_content, len(title))

        muted_setting = urwid.CheckBox(label="Muted",
                                       state=controller.model.is_muted_stream(
                                           stream_id),
                                       checked_symbol=CHECK_MARK)
        urwid.connect_signal(muted_setting, 'change', self.toggle_mute_status)
        pinned_state = controller.model.is_pinned_stream(stream_id)
        pinned_setting = urwid.CheckBox(label="Pinned to top",
                                        state=pinned_state,
                                        checked_symbol=CHECK_MARK)
        urwid.connect_signal(pinned_setting, 'change',
                             self.toggle_pinned_status)

        # Manual because calculate_table_widths does not support checkboxes.
        # Add 4 to checkbox label to accommodate the checkbox itself.
        popup_width = max(popup_width, len(muted_setting.label) + 4,
                          len(pinned_setting.label) + 4)
        self.widgets = self.make_table_with_categories(stream_info_content,
                                                       column_widths)
        self.widgets.append(muted_setting)
        self.widgets.append(pinned_setting)
        super().__init__(controller, self.widgets, 'STREAM_DESC', popup_width,
                         title)

    def toggle_mute_status(self, button: Any, new_state: bool) -> None:
        self.controller.model.toggle_stream_muted_status(self.stream_id)

    def toggle_pinned_status(self, button: Any, new_state: bool) -> None:
        self.controller.model.toggle_stream_pinned_status(self.stream_id)

    def keypress(self, size: urwid_Size, key: str) -> str:
        if is_command_key('STREAM_MEMBERS', key):
            self.controller.show_stream_members(stream_id=self.stream_id)
        return super().keypress(size, key)


class StreamMembersView(PopUpView):
    def __init__(self, controller: Any, stream_id: int) -> None:
        self.stream_id = stream_id
        self.controller = controller
        model = controller.model

        user_ids = model.get_other_subscribers_in_stream(stream_id=stream_id)
        user_names = [model.user_name_from_id(id) for id in user_ids]
        sorted_user_names = sorted(user_names)
        sorted_user_names.insert(0, model.user_full_name)
        title = 'Stream Members (up/down scrolls)'

        stream_users_content = [('', [
            (name, '') for name in sorted_user_names]
        )]
        popup_width, column_width = self.calculate_table_widths(
            stream_users_content, len(title)
        )
        widgets = self.make_table_with_categories(stream_users_content,
                                                  column_width)

        super().__init__(controller, widgets, 'STREAM_DESC', popup_width,
                         title)

    def keypress(self, size: urwid_Size, key: str) -> str:
        if(is_command_key('GO_BACK', key)
                or is_command_key('STREAM_MEMBERS', key)):
            self.controller.show_stream_info(stream_id=self.stream_id)
            return key
        return super().keypress(size, key)


class MsgInfoView(PopUpView):
    def __init__(self, controller: Any, msg: Message, title: str,
                 message_links: 'OrderedDict[str, Tuple[str, int, bool]]',
                 time_mentions: List[Tuple[str, str]],
                 ) -> None:
        self.msg = msg
        self.message_links = message_links
        self.time_mentions = time_mentions

        msg_info = [
            ('', [('Date & Time', time.ctime(msg['timestamp'])[:-5]),
                  ('Sender', msg['sender_full_name']),
                  ('Sender\'s Email ID', msg['sender_email'])]),
        ]
        # Only show the 'Edit History' label for edited messages.
        self.show_edit_history_label = (
            self.msg['id'] in controller.model.index['edited_messages']
            and controller.model.initial_data['realm_allow_edit_history']
        )
        if self.show_edit_history_label:
            msg_info[0][1][0] = (
                'Date & Time (Original)', time.ctime(msg['timestamp'])[:-5]
            )

            keys = ', '.join(map(repr, keys_for_command('EDIT_HISTORY')))
            msg_info[0][1].append(
                ('Edit History', 'Press {} to view'.format(keys))
            )
        # Render the category using the existing table methods if links exist.
        if message_links:
            msg_info.append(('Message Links', []))
        if time_mentions:
            msg_info.append(('Time mentions', time_mentions))
        if msg['reactions']:
            reactions = sorted(
                (reaction['emoji_name'], reaction['user']['full_name'])
                for reaction in msg['reactions']
            )
            grouped_reactions = dict()  # type: Dict[str, str]
            for reaction, user in reactions:
                if reaction in grouped_reactions:
                    grouped_reactions[reaction] += '\n{}'.format(user)
                else:
                    grouped_reactions[reaction] = user
            msg_info.append(('Reactions', list(grouped_reactions.items())))

        popup_width, column_widths = self.calculate_table_widths(msg_info,
                                                                 len(title))
        widgets = self.make_table_with_categories(msg_info, column_widths)

        if message_links:
            message_link_widgets = []
            message_link_width = 0
            for index, link in enumerate(message_links):
                text, link_index, _ = message_links[link]
                caption = '{}: {}\n{}'.format(link_index, text, link)
                message_link_width = max(
                    message_link_width,
                    len(max(caption.split('\n'), key=len))
                )

                display_attr = None if index % 2 else 'popup_contrast'
                message_link_widgets.append(
                    MessageLinkButton(controller, caption, link, display_attr)
                )

            # slice_index = Number of labels before message links + 1 newline
            #               + 1 'Message Links' category label.
            slice_index = len(msg_info[0][1]) + 2
            widgets = (widgets[:slice_index] + message_link_widgets
                       + widgets[slice_index:])
            popup_width = max(popup_width, message_link_width)

        super().__init__(controller, widgets, 'MSG_INFO', popup_width, title)

    def keypress(self, size: urwid_Size, key: str) -> str:
        if (is_command_key('EDIT_HISTORY', key)
                and self.show_edit_history_label):
            self.controller.show_edit_history(
                message=self.msg,
                message_links=self.message_links,
                time_mentions=self.time_mentions,
            )
        return super().keypress(size, key)


class EditModeView(PopUpView):
    def __init__(self, controller: Any, button: Any):
        self.edit_mode_button = button
        self.widgets = []  # type: List[urwid.RadioButton]

        for mode in ['change_one', 'change_later', 'change_all']:
            self.add_radio_button(mode)
        super().__init__(controller, self.widgets, 'ENTER', 62,
                         'Topic edit propagation mode')
        # Set cursor to marked checkbox.
        for i in range(len(self.widgets)):
            if self.widgets[i].state:
                self.set_focus(i)

    def set_selected_mode(self, button: Any, new_state: bool,
                          mode: str) -> None:
        if new_state:
            self.edit_mode_button.set_selected_mode(mode)

    def add_radio_button(self, mode: str) -> None:
        state = mode == self.edit_mode_button.mode
        radio_button = urwid.RadioButton(self.widgets,
                                         edit_mode_captions[mode],
                                         state=state)
        urwid.connect_signal(radio_button, 'change', self.set_selected_mode,
                             mode)

    def keypress(self, size: urwid_Size, key: str) -> str:
        # Use space to select radio-button and exit popup too.
        if key == ' ':
            key = 'enter'
        return super().keypress(size, key)


EditHistoryTag = Literal['(Current Version)', '(Original Version)', '']


class EditHistoryView(PopUpView):
    def __init__(self, controller: Any, message: Message,
                 message_links: 'OrderedDict[str, Tuple[str, int, bool]]',
                 time_mentions: List[Tuple[str, str]],
                 title: str) -> None:
        self.controller = controller
        self.message = message
        self.message_links = message_links
        self.time_mentions = time_mentions
        width = 64
        widgets = []  # type: List[Any]

        message_history = self.controller.model.fetch_message_history(
            message_id=self.message['id'],
        )
        for index, snapshot in enumerate(message_history):
            if len(widgets) > 0:  # Separate edit blocks with newline.
                widgets.append(urwid.Text(''))

            tag = ''  # type: EditHistoryTag
            if index == 0:
                tag = '(Original Version)'
            elif index == len(message_history) - 1:
                tag = '(Current Version)'

            widgets.append(self._make_edit_block(snapshot, tag))

        if not widgets:
            feedback = [
                'Could not find any message history. See ',
                ('msg_bold', 'footer'),
                ' for the error message.',
            ]
            widgets.append(urwid.Text(feedback, align='center'))

        super().__init__(controller, widgets, 'MSG_INFO', width, title)

    def _make_edit_block(self, snapshot: Dict[str, Any],
                         tag: EditHistoryTag) -> Any:
        content = snapshot['content']
        topic = snapshot['topic']
        timestamp = time.ctime(snapshot['timestamp'])[:-5]

        user_id = snapshot.get('user_id')
        if user_id:
            author_name = self.controller.model.user_name_from_id(user_id)
        else:
            author_name = 'Author N/A'
        author_prefix = self._get_author_prefix(snapshot, tag)
        author = '{} by {}'.format(author_prefix, author_name)

        header = [
            urwid.Text(('edit_topic', topic)),
            # 18 = max(EditHistoryTag).
            (18, urwid.Text(('edit_tag', tag), align='right')),
        ]
        subheader = [
            urwid.Text(('edit_author', author)),
            # 19 = len(timestamp).
            (19, urwid.Text(('edit_time', timestamp), align='right')),
        ]

        edit_block = [
            urwid.AttrWrap(
                urwid.Columns(header, dividechars=2),
                'popup_contrast',
            ),
            urwid.Columns(subheader, dividechars=2),
            urwid.Text(content),
        ]
        return urwid.Pile(edit_block)

    @staticmethod
    def _get_author_prefix(snapshot: Dict[str, Any],
                           tag: EditHistoryTag) -> str:
        if tag == '(Original Version)':
            return 'Posted'

        # NOTE: The false alarm bit in the subsequent code block is a
        # workaround for the inconsistency in the message history.
        content = snapshot['content']
        topic = snapshot['topic']

        false_alarm_content = content == snapshot.get('prev_content')
        false_alarm_topic = topic == snapshot.get('prev_topic')
        if ('prev_topic' in snapshot and 'prev_content' in snapshot
                and not(false_alarm_content or false_alarm_topic)):
            author_prefix = 'Content & Topic edited'
        elif 'prev_content' in snapshot and not false_alarm_content:
            author_prefix = 'Content edited'
        elif 'prev_topic' in snapshot and not false_alarm_topic:
            author_prefix = 'Topic edited'
        else:
            author_prefix = 'Edited but no changes made'

        return author_prefix

    def keypress(self, size: urwid_Size, key: str) -> str:
        if (is_command_key('GO_BACK', key)
                or is_command_key('EDIT_HISTORY', key)):
            self.controller.show_msg_info(
                msg=self.message,
                message_links=self.message_links,
                time_mentions=self.time_mentions,
            )
            return key
        return super().keypress(size, key)



import os
import platform
import subprocess
import time
from collections import OrderedDict, defaultdict
from functools import wraps
from itertools import chain, combinations
from re import ASCII, MULTILINE, findall, match
from threading import Thread
from typing import (
    Any,
    Callable,
    DefaultDict,
    Dict,
    FrozenSet,
    Iterable,
    List,
    Set,
    Tuple,
    TypeVar,
    Union,
)
from urllib.parse import unquote

import lxml.html
from typing_extensions import Literal, TypedDict


MACOS = platform.system() == "Darwin"
LINUX = platform.system() == "Linux"
WSL = 'microsoft' in platform.release().lower()

StreamData = TypedDict('StreamData', {
    'name': str,
    'id': int,
    'color': str,
    'invite_only': bool,
    'description': str,
})

EmojiData = TypedDict('EmojiData', {
    'code': str,
    'type': Literal['realm_emoji', 'unicode_emoji',
                    'zulip_extra_emoji'],
})

NamedEmojiData = Dict[str, EmojiData]

Message = TypedDict('Message', {
    'id': int,
    'sender_id': int,
    'content': str,
    'recipient_id': int,
    'timestamp': int,
    'client': str,
    'subject': str,  # Only for stream msgs.
    'topic_links': List[str],
    'is_me_message': bool,
    'reactions': List[Dict[str, Any]],
    'submessages': List[Dict[str, Any]],
    'flags': List[str],
    'sender_full_name': str,
    'sender_short_name': str,
    'sender_email': str,
    'sender_realm_str': str,
    'display_recipient': Any,
    'type': str,
    'stream_id': int,  # Only for stream msgs.
    'avatar_url': str,
    'content_type': str,
    'match_content': str,  # If keyword search specified in narrow params.
    'match_subject': str,  # If keyword search specified in narrow params.
  }, total=False)

Index = TypedDict('Index', {
    'pointer': Dict[str, Union[int, Set[None]]],  # narrow_str, message_id
    # Various sets of downloaded message ids (all, starred, ...)
    'all_msg_ids': Set[int],
    'starred_msg_ids': Set[int],
    'mentioned_msg_ids': Set[int],
    'private_msg_ids': Set[int],
    'private_msg_ids_by_user_ids': Dict[FrozenSet[int], Set[int]],
    'stream_msg_ids_by_stream_id': Dict[int, Set[int]],
    'topic_msg_ids': Dict[int, Dict[str, Set[int]]],
    # Extra cached information
    'edited_messages': Set[int],  # {message_ids, ...}
    'topics': Dict[int, List[str]],  # {topic names, ...}
    'search': Set[int],  # {message_id, ...}
    # Downloaded message data
    'messages': Dict[int, Message],  # message_id: Message
})

initial_index = Index(
    pointer=defaultdict(set),
    all_msg_ids=set(),
    starred_msg_ids=set(),
    mentioned_msg_ids=set(),
    private_msg_ids=set(),
    private_msg_ids_by_user_ids=defaultdict(set),
    stream_msg_ids_by_stream_id=defaultdict(set),
    topic_msg_ids=defaultdict(dict),
    edited_messages=set(),
    topics=defaultdict(list),
    search=set(),
    # mypy bug: https://github.com/python/mypy/issues/7217
    messages=defaultdict(lambda: Message()),
)


UnreadCounts = TypedDict('UnreadCounts', {
    'all_msg': int,
    'all_pms': int,
    'all_mentions': int,
    'unread_topics': Dict[Tuple[int, str], int],  # stream_id, topic
    'unread_pms': Dict[int, int],  # sender_id
    'unread_huddles': Dict[FrozenSet[int], int],  # Group pms
    'streams': Dict[int, int],  # stream_id
})


edit_mode_captions = {
    'change_one': 'Change only this message topic',
    'change_later': 'Also change later messages to this topic',
    'change_all': 'Also change previous and following messages to this topic',
}


def asynch(func: Callable[..., None]) -> Callable[..., None]:
    """
    Decorator for executing a function in a separate :class:`threading.Thread`.
    """
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        # If calling when pytest is running simply return the function
        # to avoid running in asynch mode.
        if os.environ.get("PYTEST_CURRENT_TEST"):
            return func(*args, **kwargs)

        thread = Thread(target=func, args=args, kwargs=kwargs)
        thread.daemon = True
        return thread.start()
    return wrapper


def _set_count_in_model(new_count: int, changed_messages: List[Message],
                        unread_counts: UnreadCounts) -> None:
    """
        This function doesn't explicitly set counts in model,
        but updates `unread_counts` (which can update the model
        if it's passed in, but is not tied to it).
    """
    # broader unread counts (for all_*) are updated
    # later conditionally in _set_count_in_view.
    KeyT = TypeVar('KeyT')

    def update_unreads(unreads: Dict[KeyT, int], key: KeyT) -> None:
        if key in unreads:
            unreads[key] += new_count
            if unreads[key] == 0:
                unreads.pop(key)
        elif new_count == 1:
            unreads[key] = new_count

    for message in changed_messages:
        if message['type'] == 'stream':
            stream_id = message['stream_id']
            update_unreads(unread_counts['unread_topics'],
                           (stream_id, message['subject']))
            update_unreads(unread_counts['streams'], stream_id)
        # self-pm has only one display_recipient
        # 1-1 pms have 2 display_recipient
        elif len(message['display_recipient']) <= 2:
            update_unreads(unread_counts['unread_pms'], message['sender_id'])
        else:  # If it's a group pm
            update_unreads(unread_counts['unread_huddles'],
                           frozenset(recipient['id'] for recipient
                                     in message['display_recipient']))


def _set_count_in_view(controller: Any, new_count: int,
                       changed_messages: List[Message],
                       unread_counts: UnreadCounts) -> None:
    """
        This function for the most part contains the logic for setting the
        count in the UI buttons. The later buttons (all_msg, all_pms)
        additionally set the current count in the model and make use of the
        same in the UI.
    """
    stream_buttons_log = controller.view.stream_w.log
    is_open_topic_view = controller.view.left_panel.is_in_topic_view
    if is_open_topic_view:
        topic_buttons_log = controller.view.topic_w.log
        toggled_stream_id = controller.view.topic_w.stream_button.stream_id
    user_buttons_log = controller.view.user_w.log
    all_msg = controller.view.home_button
    all_pm = controller.view.pm_button
    all_mentioned = controller.view.mentioned_button
    for message in changed_messages:
        user_id = message['sender_id']

        # If we sent this message, don't increase the count
        if user_id == controller.model.user_id:
            continue

        msg_type = message['type']
        add_to_counts = True
        if 'mentioned' in message['flags']:
            unread_counts['all_mentions'] += new_count
            all_mentioned.update_count(unread_counts['all_mentions'])

        if msg_type == 'stream':
            stream_id = message['stream_id']
            msg_topic = message['subject']
            if controller.model.is_muted_stream(stream_id):
                add_to_counts = False  # if muted, don't add to eg. all_msg
            else:
                for stream_button in stream_buttons_log:
                    if stream_button.stream_id == stream_id:
                        stream_button.update_count(stream_button.count
                                                   + new_count)
                        break
            # FIXME: Update unread_counts['unread_topics']?
            if controller.model.is_muted_topic(stream_id, msg_topic):
                add_to_counts = False
            if is_open_topic_view and stream_id == toggled_stream_id:
                # If topic_view is open for incoming messages's stream,
                # We update the respective TopicButton count accordingly.
                for topic_button in topic_buttons_log:
                    if topic_button.topic_name == msg_topic:
                        topic_button.update_count(topic_button.count
                                                  + new_count)
        else:
            for user_button in user_buttons_log:
                if user_button.user_id == user_id:
                    user_button.update_count(user_button.count + new_count)
                    break
            unread_counts['all_pms'] += new_count
            all_pm.update_count(unread_counts['all_pms'])

        if add_to_counts:
            unread_counts['all_msg'] += new_count
            all_msg.update_count(unread_counts['all_msg'])


def set_count(id_list: List[int], controller: Any, new_count: int) -> None:
    # This method applies new_count for 'new message' (1) or 'read' (-1)
    # (we could ensure this in a different way by a different type)
    assert new_count == 1 or new_count == -1
    messages = controller.model.index['messages']
    unread_counts = controller.model.unread_counts  # type: UnreadCounts
    changed_messages = [messages[id] for id in id_list]
    _set_count_in_model(new_count, changed_messages, unread_counts)

    # if view is not yet loaded. Usually the case when first message is read.
    while not hasattr(controller, 'view'):
        time.sleep(0.1)

    _set_count_in_view(controller, new_count, changed_messages, unread_counts)

    while not hasattr(controller, 'loop'):
        time.sleep(0.1)
    controller.update_screen()


def index_messages(messages: List[Message],
                   model: Any,
                   index: Index) -> Index:
    """
    STRUCTURE OF INDEX
    {
        'pointer': {
            '[]': 30  # str(ZulipModel.narrow)
            '[["stream", "verona"]]': 32,
            ...
        }
        'topic_msg_ids': {
            123: {    # stream_id
                'topic name': {
                    51234,  # message id
                    56454,
                    ...
                }
        },
        'private_msg_ids_by_user_ids': {
            (3, 7): {  # user_ids frozenset
                51234,
                56454,
                ...
            },
            (1, 2, 3, 4): {  # multiple recipients
                12345,
                32553,
            }
        },
        'topics': {
            123: [    # stread_id
                'Denmark2', # topic name
                'Verona2',
                ....
            ]
        },
        'all_msg_ids': {
            14231,
            23423,
            ...
        },
        'private_msg_ids': {
            22334,
            23423,
            ...
        },
        'mentioned_msg_ids': {
            14423,
            33234,
            ...
        },
        'stream_msg_ids_by_stream_id': {
            123: {
                53434,
                36435,
                ...
            }
            234: {
                23423,
                23423,
                ...
            }
        },
        'edited_messages':{
            51234,
            23423,
            ...
        },
        'search': {
            13242,
            23423,
            23423,
            ...
        },
        'messages': {
            # all the messages mapped to their id
            # for easy retrieval of message from id
            45645: {  # PRIVATE
                'id': 4290,
                'timestamp': 1521817473,
                'content': 'Hi @**Cordelia Lear**',
                'sender_full_name': 'Iago',
                'flags': [],
                'sender_short_name': 'iago',
                'sender_email': 'iago@zulip.com',
                'subject': '',
                'subject_links': [],
                'sender_id': 73,
                'type': 'private',
                'recipient_id': 124,
                'reactions': [],
                'display_recipient': [
                    {
                        'email': 'ZOE@zulip.com',
                        'id': 70,
                        'full_name': 'Zoe',
                    }, {
                        'email': 'cordelia@zulip.com',
                        'id': 71,
                        'full_name': 'Cordelia Lear',
                    }, {
                        'email': 'hamlet@zulip.com',
                        'id': 72,
                        'full_name': 'King Hamlet',
                    }, {
                        'email': 'iago@zulip.com',
                        'id': 73,
                        'full_name': 'Iago',
                    }
                ]
            },
            45645: {  # STREAM
                'timestamp': 1521863062,
                'sender_id': 72,
                'sender_full_name': 'King Hamlet',
                'recipient_id': 119,
                'content': 'https://github.com/zulip/zulip-terminal',
                'type': 'stream',
                'sender_email': 'hamlet@zulip.com',
                'id': 4298,
                'display_recipient': 'Verona',
                'flags': [],
                'reactions': [],
                'subject': 'Verona2',
                'stream_id': 32,
            },
        },
    }
    """
    narrow = model.narrow
    for msg in messages:

        if 'edit_history' in msg.keys():
            index['edited_messages'].add(msg['id'])

        index['messages'][msg['id']] = msg
        if not narrow:
            index['all_msg_ids'].add(msg['id'])

        elif model.is_search_narrow():
            index['search'].add(msg['id'])
            continue

        if len(narrow) == 1:

            if narrow[0][1] == 'starred':
                if 'starred' in msg['flags']:
                    index['starred_msg_ids'].add(msg['id'])

            if narrow[0][1] == 'mentioned':
                if 'mentioned' in msg['flags']:
                    index['mentioned_msg_ids'].add(msg['id'])

            if msg['type'] == 'private':
                index['private_msg_ids'].add(msg['id'])
                recipients = frozenset({
                    recipient['id'] for recipient in msg['display_recipient']
                })

                if narrow[0][0] == 'pm_with':
                    narrow_emails = ([model.user_dict[email]['user_id']
                                      for email in narrow[0][1].split(', ')]
                                     + [model.user_id])
                    if recipients == frozenset(narrow_emails):
                        (index['private_msg_ids_by_user_ids'][recipients]
                         .add(msg['id']))

            if msg['type'] == 'stream' and msg['stream_id'] == model.stream_id:
                (index['stream_msg_ids_by_stream_id'][msg['stream_id']]
                 .add(msg['id']))

        if (msg['type'] == 'stream' and len(narrow) == 2
                and narrow[1][1] == msg['subject']):
            topics_in_stream = index['topic_msg_ids'][msg['stream_id']]
            if not topics_in_stream.get(msg['subject']):
                topics_in_stream[msg['subject']] = set()
            topics_in_stream[msg['subject']].add(msg['id'])

    return index


def classify_unread_counts(model: Any) -> UnreadCounts:
    # TODO: support group pms
    unread_msg_counts = model.initial_data['unread_msgs']

    unread_counts = UnreadCounts(
        all_msg=0,
        all_pms=0,
        all_mentions=0,
        unread_topics=dict(),
        unread_pms=dict(),
        unread_huddles=dict(),
        streams=defaultdict(int),
    )

    mentions_count = len(unread_msg_counts['mentions'])
    unread_counts['all_mentions'] += mentions_count

    for pm in unread_msg_counts['pms']:
        count = len(pm['unread_message_ids'])
        unread_counts['unread_pms'][pm['sender_id']] = count
        unread_counts['all_msg'] += count
        unread_counts['all_pms'] += count

    for stream in unread_msg_counts['streams']:
        count = len(stream['unread_message_ids'])
        stream_id = stream['stream_id']
        # unsubscribed streams may be in raw unreads, but are not tracked
        if not model.is_user_subscribed_to_stream(stream_id):
            continue
        if model.is_muted_topic(stream_id, stream['topic']):
            continue
        stream_topic = (stream_id, stream['topic'])
        unread_counts['unread_topics'][stream_topic] = count
        if not unread_counts['streams'].get(stream_id):
            unread_counts['streams'][stream_id] = count
        else:
            unread_counts['streams'][stream_id] += count
        if stream_id not in model.muted_streams:
            unread_counts['all_msg'] += count

    # store unread count of group pms in `unread_huddles`
    for group_pm in unread_msg_counts['huddles']:
        count = len(group_pm['unread_message_ids'])
        user_ids = group_pm['user_ids_string'].split(',')
        user_ids = frozenset(map(int, user_ids))
        unread_counts['unread_huddles'][user_ids] = count
        unread_counts['all_msg'] += count
        unread_counts['all_pms'] += count

    return unread_counts


def match_user(user: Any, text: str) -> bool:
    """
    Matches if the user full name, last name or email matches
    with `text` or not.
    """
    full_name = user['full_name'].lower()
    keywords = full_name.split()
    # adding full_name helps in further narrowing down the right user.
    keywords.append(full_name)
    keywords.append(user['email'].lower())
    for keyword in keywords:
        if keyword.startswith(text.lower()):
            return True
    return False


def match_emoji(emoji: str, text: str) -> bool:
    """
    True if the emoji matches with `text` (case insensitive),
    False otherwise.
    """
    return emoji.lower().startswith(text.lower())


def match_topics(topic_names: List[str], search_text: str) -> List[str]:
    return [name for name in topic_names
            if name.lower().startswith(search_text.lower())]


DataT = TypeVar('DataT')


def match_stream(data: List[Tuple[DataT, str]], search_text: str,
                 pinned_streams: List[StreamData]
                 ) -> Tuple[List[DataT], List[str]]:
    """
    Returns a list of DataT (streams) and a list of their corresponding names
    whose words match with the 'text' in the following order:
    * 1st-word startswith match > 2nd-word startswith match > ... (pinned)
    * 1st-word startswith match > 2nd-word startswith match > ... (unpinned)

    Note: This function expects `data` to be sorted, in a non-decreasing
    order, and ordered by their pinning status.
    """
    pinned_stream_names = [stream['name'] for stream in pinned_streams]

    # Assert that the data is sorted, in a non-decreasing order, and ordered by
    # their pinning status.
    assert data == sorted(sorted(data, key=lambda data: data[1].lower()),
                          key=lambda data: data[1] in pinned_stream_names,
                          reverse=True)

    delimiters = '-_/'
    trans = str.maketrans(delimiters, len(delimiters) * ' ')
    stream_splits = [
        ((datum, [stream_name] + stream_name.translate(trans).split()[1:]))
        for datum, stream_name in data
    ]

    matches = OrderedDict([
        ('pinned', defaultdict(list)),
        ('unpinned', defaultdict(list)),
    ])  # type: OrderedDict[str, DefaultDict[int, List[Tuple[DataT, str]]]]

    for datum, splits in stream_splits:
        stream_name = splits[0]
        kind = 'pinned' if stream_name in pinned_stream_names else 'unpinned'
        for match_position, word in enumerate(splits):
            if word.lower().startswith(search_text.lower()):
                matches[kind][match_position].append((datum, stream_name))

    ordered_matches = []
    ordered_names = []
    for matched_data in matches.values():
        if not matched_data:
            continue
        for match_position in range(max(matched_data.keys()) + 1):
            for datum, name in matched_data.get(match_position, []):
                if datum not in ordered_matches:
                    ordered_matches.append(datum)
                    ordered_names.append(name)
    return ordered_matches, ordered_names


def match_group(group_name: str, text: str) -> bool:
    """
    True if any group name matches with `text` (case insensitive),
    False otherwise.
    """
    return group_name.lower().startswith(text.lower())


def format_string(names: List[str], wrapping_text: str) -> List[str]:
    """
    Wrap a list of names using the wrapping characters for typeahead
    """
    return [wrapping_text.format(name) for name in names]


def powerset(iterable: Iterable[Any],
             map_func: Callable[[Any], Any]=set) -> List[Any]:
    """
    >> powerset([1,2,3])
    returns: [set(), {1}, {2}, {3}, {1, 2}, {1, 3}, {2, 3}, {1, 2, 3}]"
    """
    s = list(iterable)
    powerset = chain.from_iterable(combinations(s, r)
                                   for r in range(len(s) + 1))
    return list(map(map_func, list(powerset)))


def canonicalize_color(color: str) -> str:
    """
    Given a color of the format '#xxxxxx' or '#xxx', produces one of the
    format '#xxx'. Always produces lowercase hex digits.
    """
    if match('^#[0-9A-Fa-f]{6}$', color, ASCII) is not None:
        # '#xxxxxx' color, stored by current zulip server
        return (color[:2] + color[3] + color[5]).lower()
    elif match('^#[0-9A-Fa-f]{3}$', color, ASCII) is not None:
        # '#xxx' color, which may be stored by the zulip server <= 2.0.0
        # Potentially later versions too
        return color.lower()
    else:
        raise ValueError('Unknown format for color "{}"'.format(color))


def notify(title: str, html_text: str) -> str:
    document = lxml.html.document_fromstring(html_text)
    text = document.text_content()

    command_list = None
    if MACOS:
        command_list = [
            "osascript",
            "-e", "on run(argv)",
            "-e", "return display notification item 1 of argv with title "
            'item 2 of argv sound name "ZT_NOTIFICATION_SOUND"',
            "-e", "end",
            "--", text, title
        ]
    elif LINUX:
        command_list = ["notify-send", "--", title, text]

    if command_list is not None:
        try:
            subprocess.run(command_list, stdout=subprocess.DEVNULL,
                           stderr=subprocess.DEVNULL)
        except FileNotFoundError:
            # This likely means the notification command could not be found
            return command_list[0]
    return ""


def display_error_if_present(response: Dict[str, Any], controller: Any
                             ) -> None:
    if response['result'] == 'error' and hasattr(controller, 'view'):
        controller.view.set_footer_text(response['msg'], 3)


def hash_util_decode(string: str) -> str:
    """
    Returns a decoded string given a hash_util_encode() [present in
    zulip/zulip's zerver/lib/url_encoding.py] encoded string.
    """
    # Acknowledge custom string replacements in zulip/zulip's
    # zerver/lib/url_encoding.py before unquote.
    return unquote(string.replace('.', '%'))


def get_unused_fence(content: str) -> str:
    """
    Generates fence for quoted-message based on regex pattern
    of continuous back-ticks. Referred and translated from
    zulip/static/shared/js/fenced_code.js.
    """
    fence_length_regex = '^ {0,3}(`{3,})'
    max_length_fence = 3

    matches = findall(fence_length_regex, content,
                      flags=MULTILINE)
    if len(matches) != 0:
        max_length_fence = max(max_length_fence,
                               len(max(matches, key=len)) + 1)

    return '`' * max_length_fence



from typing import Dict, List, Optional, Tuple


ThemeSpec = List[Tuple[Optional[str], ...]]

# The keys in required_styles specify what styles are necessary for a theme to
# be complete, while the values are those used to style each element in
# monochrome (1-bit) mode - independently of the specified theme

required_styles = {  # style-name: monochrome-bit-depth-style
    None: '',
    'selected': 'standout',
    'msg_selected': 'standout',
    'header': 'bold',
    'general_narrow': 'standout',
    'general_bar': '',
    'name': '',
    'unread': 'strikethrough',
    'user_active': 'bold',
    'user_idle': '',
    'user_offline': '',
    'user_inactive': '',
    'title': 'bold',
    'time': '',
    'bar': 'standout',
    'popup_contrast': 'standout',
    'msg_emoji': 'bold',
    'reaction': 'bold',
    'reaction_mine': 'standout',
    'msg_mention': 'bold',
    'msg_link': '',
    'msg_link_index': 'bold',
    'msg_quote': 'underline',
    'msg_code': 'bold',
    'msg_bold': 'bold',
    'msg_time': 'bold',
    'footer': 'standout',
    'starred': 'bold',
    'popup_category': 'bold',
    'unread_count': 'bold',
    'filter_results': 'bold',
    'edit_topic': 'standout',
    'edit_tag': 'standout',
    'edit_author': 'bold',
    'edit_time': 'bold',
    'popup_border': 'bold',
    'area:help': 'standout',
    'area:msg': 'standout',
    'area:stream': 'standout',
    'area:error': 'standout',
}


# 256-color base names for default theme (bold values added below)
DEF_base = dict(
    dark_red='#a00',
    brown='#880',
    dark_blue='#24a',
    dark_cyan='#088',
    dark_gray='#666',
    light_gray='#ccc',
    light_red='#f00',
    light_green='#0f0',
    dark_green='#080',
    yellow='#ff0',
    light_blue='#28d',
    light_magenta='#c8f',
    white='#fff',
    black='g19',
)

DEF = dict(DEF_base,
           **{color + ':bold': code + ', bold'
              for color, code in DEF_base.items()})

# Colors used in gruvbox-256
# See https://github.com/morhetz/gruvbox/blob/master/colors/gruvbox.vim
BLACK = 'h234'  # dark0_hard
WHITE = 'h246'  # light4_256
WHITEBOLD = '%s, bold' % WHITE
DARKBLUE = 'h24'  # faded_blue
DARKRED = 'h88'  # faded_red
LIGHTBLUE = 'h109'  # bright_blue
LIGHTBLUEBOLD = '%s, bold' % LIGHTBLUE
YELLOW = 'h172'  # neutral_yellow
YELLOWBOLD = '%s, bold' % YELLOW
LIGHTGREEN = 'h142'  # bright_green
LIGHTRED = 'h167'  # bright_red
LIGHTREDBOLD = '%s, bold' % LIGHTRED
GRAY = 'h244'  # gray_244
LIGHTMAGENTA = 'h132'  # neutral_purple
LIGHTMAGENTABOLD = '%s, bold' % LIGHTMAGENTA

THEME_ALIASES = {
    'default': 'zt_dark',
    'gruvbox': 'gruvbox_dark',
    'light': 'zt_light',
    'blue': 'zt_blue',
}

THEMES = {
    'zt_dark': [
        (None,             'white',                   'black',
         None,             DEF['white'],              DEF['black']),
        ('selected',       'white',                   'dark blue',
         None,             DEF['white'],              DEF['dark_blue']),
        ('msg_selected',   'white',                   'dark blue',
         None,             DEF['white'],              DEF['dark_blue']),
        ('header',         'dark cyan',               'dark blue',
         None,             DEF['dark_cyan'],          DEF['dark_blue']),
        ('general_narrow', 'white',                   'dark blue',
         None,             DEF['white'],              DEF['dark_blue']),
        ('general_bar',    'white',                   'black',
         None,             DEF['white'],              DEF['black']),
        ('name',           'yellow, bold',            'black',
         None,             DEF['yellow:bold'],        DEF['black']),
        ('unread',         'dark blue',               'black',
         None,             DEF['dark_blue'],          DEF['black']),
        ('user_active',    'light green',             'black',
         None,             DEF['light_green'],        DEF['black']),
        ('user_idle',      'yellow',                  'black',
         None,             DEF['yellow'],             DEF['black']),
        ('user_offline',   'white',                   'black',
         None,             DEF['white'],              DEF['black']),
        ('user_inactive',  'white',                   'black',
         None,             DEF['white'],              DEF['black']),
        ('title',          'white, bold',             'black',
         None,             DEF['white:bold'],         DEF['black']),
        ('time',           'light blue',              'black',
         None,             DEF['light_blue'],         DEF['black']),
        ('bar',            'white',                   'dark gray',
         None,             DEF['white'],              DEF['dark_gray']),
        ('popup_contrast', 'white',                   'dark gray',
         None,             DEF['white'],              DEF['dark_gray']),
        ('msg_emoji',      'light magenta',           'black',
         None,             DEF['light_magenta'],      DEF['black']),
        ('reaction',       'light magenta, bold',     'black',
         None,             DEF['light_magenta:bold'], DEF['black']),
        ('reaction_mine',  'black',                   'light magenta',
         None,             DEF['black'],              DEF['light_magenta']),
        ('msg_mention',    'light red, bold',         'black',
         None,             DEF['light_red:bold'],     DEF['black']),
        ('msg_link',       'light blue',              'black',
         None,             DEF['light_blue'],         DEF['black']),
        ('msg_link_index', 'light blue, bold',        'black',
         None,             DEF['light_blue:bold'],    DEF['black']),
        ('msg_quote',      'brown',                   'black',
         None,             DEF['brown'],              DEF['black']),
        ('msg_code',       'black',                   'white',
         None,             DEF['black'],              DEF['white']),
        ('msg_bold',       'white, bold',             'black',
         None,             DEF['white:bold'],         DEF['black']),
        ('msg_time',       'black',                   'white',
         None,             DEF['black'],              DEF['white']),
        ('footer',         'white',                   'dark red',
         None,             DEF['white'],              DEF['dark_red']),
        ('starred',        'light red, bold',         'black',
         None,             DEF['light_red:bold'],     DEF['black']),
        ('popup_category', 'light blue, bold',        'black',
         None,             DEF['light_blue:bold'],    DEF['black']),
        ('unread_count',   'yellow',                  'black',
         None,             DEF['yellow'],             DEF['black']),
        ('table_head',     'white, bold',             'black',
         None,             DEF['white:bold'],         DEF['black']),
        ('filter_results', 'white',                   'dark green',
         None,             DEF['white'],              DEF['dark_green']),
        ('edit_topic',     'white',                   'dark gray',
         None,             DEF['white'],              DEF['dark_gray']),
        ('edit_tag',       'white',                   'dark gray',
         None,             DEF['white'],              DEF['dark_gray']),
        ('edit_author',    'yellow',                  'black',
         None,             DEF['yellow'],             DEF['black']),
        ('edit_time',      'light blue',              'black',
         None,             DEF['light_blue'],         DEF['black']),
        ('current_user',   'white',                   'black',
         None,             DEF['white'],              DEF['black']),
        ('muted',          'light blue',              'black',
         None,             DEF['light_blue'],         DEF['black']),
        ('popup_border',   'white',                   'black',
         None,             DEF['white'],              DEF['black']),
        ('area:help',      'white',                   'dark green',
         None,             DEF['white'],              DEF['dark_green']),
        ('area:msg',       'white',                   'brown',
         None,             DEF['white'],              DEF['brown']),
        ('area:stream',    'white',                   'dark cyan',
         None,             DEF['white'],              DEF['dark_cyan']),
        ('area:error',     'white',                   'dark red',
         None,             DEF['white'],               DEF['dark_red']),
    ],
    'gruvbox_dark': [
        # default colorscheme on 16 colors, gruvbox colorscheme
        # on 256 colors
        (None,             'white',           'black',
         None,             WHITE,             BLACK),
        ('selected',       'black',           'white',
         None,             BLACK,             WHITE),
        ('msg_selected',   'black',           'white',
         None,             BLACK,             WHITE),
        ('header',         'dark cyan',       'dark blue',
         None,             'dark cyan',       DARKBLUE),
        ('general_narrow', 'white',           'dark blue',
         None,             WHITE,             DARKBLUE),
        ('general_bar',    'white',           'black',
         None,             WHITE,             BLACK),
        ('name',           'yellow, bold',    'black',
         None,             YELLOWBOLD,        BLACK),
        ('unread',         'light magenta',   'black',
         None,             LIGHTMAGENTA,      BLACK),
        ('user_active',    'light green',     'black',
         None,             LIGHTGREEN,        BLACK),
        ('user_idle',      'yellow',          'black',
         None,             YELLOW,            BLACK),
        ('user_offline',   'white',           'black',
         None,             WHITE,             BLACK),
        ('user_inactive',  'white',           'black',
         None,             WHITE,             BLACK),
        ('title',          'white, bold',     'black',
         None,             WHITEBOLD,         BLACK),
        ('time',           'light blue',      'black',
         None,             LIGHTBLUE,         BLACK),
        ('bar',            'white',           'dark gray',
         None,             WHITE,             GRAY),
        ('popup_contrast', 'black',           'dark gray',
         None,             BLACK,             GRAY),
        ('msg_emoji',      'light magenta',   'black',
         None,             LIGHTMAGENTA,   BLACK),
        ('reaction',       'light magenta, bold', 'black',
         None,             LIGHTMAGENTABOLD,   BLACK),
        ('reaction_mine',  'black',           'light magenta',
         None,             BLACK,             LIGHTMAGENTA),
        ('msg_mention',    'light red, bold', 'black',
         None,             LIGHTREDBOLD,      BLACK),
        ('msg_link',       'light blue',      'black',
         None,             LIGHTBLUE,         BLACK),
        ('msg_link_index', 'light blue, bold', 'black',
         None,             LIGHTBLUEBOLD,     BLACK),
        ('msg_quote',      'brown',           'black',
         None,             'brown',           BLACK),
        ('msg_code',       'black',           'white',
         None,             BLACK,             WHITE),
        ('msg_bold',       'white, bold',     'black',
         None,             WHITEBOLD,         BLACK),
        ('msg_time',       'black',           'white',
         None,             BLACK,             WHITE),
        ('footer',         'white',           'dark red',
         None,             WHITE,             DARKRED),
        ('starred',        'light red, bold', 'black',
         None,             LIGHTREDBOLD,      BLACK),
        ('popup_category', 'light blue, bold', 'black',
         None,             LIGHTBLUE,         BLACK),
        ('unread_count',   'yellow',          'black',
         None,             YELLOW,            BLACK),
        ('table_head',     'white, bold',     'black',
         None,             WHITEBOLD,         BLACK),
        ('filter_results', 'black',           'light green',
         None,             BLACK,             LIGHTGREEN),
        ('edit_topic',     'black',           'dark gray',
         None,             BLACK,             GRAY),
        ('edit_tag',       'black',           'dark gray',
         None,             BLACK,             GRAY),
        ('edit_author',    'yellow',          'black',
         None,             YELLOW,            BLACK),
        ('edit_time',      'light blue',      'black',
         None,             LIGHTBLUE,         BLACK),
        ('current_user',   'white',           'black',
         None,             WHITE,             BLACK),
        ('muted',          'light blue',      'black',
         None,             LIGHTBLUE,         BLACK),
        ('popup_border',   'white',           'black',
         None,             WHITE,             BLACK),
        ('area:help',      'black',           'light green',
         None,             BLACK,             LIGHTGREEN),
        ('area:msg',       'black',           'light red',
         None,             BLACK,             LIGHTRED),
        ('area:stream',    'black',           'light blue',
         None,             BLACK,             LIGHTBLUE),
        ('area:error',     'white',           'dark red',
         None,             WHITE,             DARKRED),
    ],
    'zt_light': [
        (None,             'black',           'white'),
        ('selected',       'black',           'light green'),
        ('msg_selected',   'black',           'light green'),
        ('header',         'white',           'dark blue'),
        ('general_narrow', 'white',           'dark blue'),
        ('general_bar',    'dark blue',       'white'),
        ('name',           'dark green',      'white'),
        ('unread',         'dark gray',       'light gray'),
        ('user_active',    'dark green',      'white'),
        ('user_idle',      'dark blue',       'white'),
        ('user_offline',   'black',           'white'),
        ('user_inactive',  'black',           'white'),
        ('title',          'white, bold',     'dark gray'),
        ('time',           'dark blue',           'white'),
        ('bar',            'white',           'dark gray'),
        ('popup_contrast', 'white',           'dark gray'),
        ('msg_emoji',      'light magenta',   'white'),
        ('reaction',       'light magenta, bold',   'white'),
        ('reaction_mine',  'white',           'light magenta'),
        ('msg_mention',    'light red, bold', 'white'),
        ('msg_link',       'dark blue',       'white'),
        ('msg_link_index', 'dark blue, bold', 'white'),
        ('msg_quote',      'black',           'brown'),
        ('msg_code',       'black',           'light gray'),
        ('msg_bold',       'white, bold',     'dark gray'),
        ('msg_time',       'white',           'dark gray'),
        ('footer',         'white',           'dark red'),
        ('starred',        'light red, bold', 'white'),
        ('popup_category', 'dark gray, bold', 'light gray'),
        ('unread_count',   'dark blue, bold', 'white'),
        ('table_head',     'black, bold',     'white'),
        ('filter_results', 'white',           'dark green'),
        ('edit_topic',     'white',           'dark gray'),
        ('edit_tag',       'white',           'dark gray'),
        ('edit_author',    'dark green',      'white'),
        ('edit_time',      'dark blue',       'white'),
        ('current_user',   'dark gray',       'white'),
        ('muted',          'dark gray',       'white'),
        ('popup_border',   'black',           'white'),
        ('area:help',      'black',           'light green'),
        ('area:stream',    'black',           'light blue'),
        ('area:msg',       'black',           'yellow'),
        ('area:error',     'black',           'light red'),
    ],
    'zt_blue': [
        (None,             'black',           'light blue'),
        ('selected',       'black',           'light gray'),
        ('msg_selected',   'black',           'light gray'),
        ('header',         'black',           'dark blue'),
        ('general_narrow', 'white',           'dark blue'),
        ('general_bar',    'dark blue',       'light blue'),
        ('name',           'dark red',        'light blue'),
        ('unread',         'light gray',      'light blue'),
        ('user_active',    'light green, bold', 'light blue'),
        ('user_idle',      'dark gray',       'light blue'),
        ('user_offline',   'black',           'light blue'),
        ('user_inactive',  'black',           'light blue'),
        ('title',          'white, bold',     'dark blue'),
        ('time',           'dark blue',       'light blue'),
        ('bar',            'white',           'dark blue'),
        ('popup_contrast', 'white',           'dark blue'),
        ('msg_emoji',      'dark magenta',   'light blue'),
        ('reaction',       'dark magenta, bold', 'light blue'),
        ('reaction_mine',  'light blue',      'dark magenta'),
        ('msg_mention',    'light red, bold', 'light blue'),
        ('msg_link',       'dark blue',       'light gray'),
        ('msg_link_index', 'dark blue, bold', 'light gray'),
        ('msg_quote',      'brown',           'dark blue'),
        ('msg_code',       'dark blue',       'white'),
        ('msg_bold',       'white, bold',     'dark blue'),
        ('msg_time',       'dark blue',       'white'),
        ('footer',         'white',           'dark red'),
        ('starred',        'light red, bold', 'light blue'),
        ('popup_category', 'light gray, bold', 'light blue'),
        ('unread_count',   'yellow',          'light blue'),
        ('table_head',     'black, bold',     'light blue'),
        ('filter_results', 'white',           'dark green'),
        ('edit_topic',     'white',           'dark blue'),
        ('edit_tag',       'white',           'dark blue'),
        ('edit_author',    'dark gray',       'light blue'),
        ('edit_time',      'dark blue',       'light blue'),
        ('current_user',   'light gray',      'light blue'),
        ('muted',          'light gray',      'light blue'),
        ('popup_border',   'white',           'light blue'),
        ('area:help',      'white',           'dark green'),
        ('area:stream',    'white',           'dark cyan'),
        ('area:msg',       'white',           'brown'),
        ('area:error',     'white',           'dark red'),
    ]
}  # type: Dict[str, ThemeSpec]


def all_themes() -> List[str]:
    return list(THEMES.keys())


def aliased_themes() -> Dict[str, str]:
    return dict(THEME_ALIASES)


def complete_and_incomplete_themes() -> Tuple[List[str], List[str]]:
    complete = {name for name, styles in THEMES.items()
                if {s[0] for s in styles}.issuperset(required_styles)}
    incomplete = list(set(THEMES) - complete)
    return sorted(list(complete)), sorted(incomplete)


def theme_with_monochrome_added(theme: ThemeSpec) -> ThemeSpec:
    updated_theme = []
    for style in theme:
        style_name = style[0]
        if style_name not in required_styles:  # incomplete theme
            continue
        mono_style = required_styles[style_name]
        if len(style) > 4:     # 256 colors+
            new_style = style[:3] + (mono_style,) + style[4:]
        elif len(style) == 4:  # 16 colors + mono (overwrite mono)
            new_style = style[:3] + (mono_style,)
        elif len(style) == 3:  # 16 colors only
            new_style = style[:3] + (mono_style,)
        else:                  # 1-to-1 mapping (same as other style)
            new_style = style
        updated_theme.append(new_style)
    return updated_theme



STREAM_MARKER_INVALID = '✗'
STREAM_MARKER_PRIVATE = 'P'
STREAM_MARKER_PUBLIC = '#'
STREAM_TOPIC_SEPARATOR = '▶'
# Used as a separator between messages and 'EDITED'
MESSAGE_CONTENT_MARKER = '▒'  # Options are '█', '▓', '▒', '░'
QUOTED_TEXT_MARKER = '░'
MESSAGE_HEADER_DIVIDER = '━'
CHECK_MARK = '✓'
APPLICATION_TITLE_BAR_LINE = '═'
PINNED_STREAMS_DIVIDER = '-'
LIST_TITLE_BAR_LINE = '━'
# NOTE: The '⏱' emoji needs an extra space while rendering. Otherwise, it
# appears to overlap its subsequent text.
TIME_MENTION_MARKER = '⏱ '  # Other tested options are: '⧗' and '⧖'.
MUTE_MARKER = 'M'
STATUS_ACTIVE = '●'
STATUS_IDLE = '◒'
STATUS_OFFLINE = '○'
STATUS_INACTIVE = '•'
